# -*- coding: utf-8 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                 Contact                     #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2013  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys
from webbrowser import open_new_tab
from urllib import urlencode

from PyQt4.QtCore import pyqtSignal
from PyQt4.QtGui import (QHBoxLayout, QVBoxLayout, QCheckBox, QPushButton, QDialog,
                         QTextEdit, QLineEdit, QLabel, QComboBox, QGroupBox,
                         QMessageBox, QWidget,)
from .. import param
from ..pylib import path2
from ..pylib.bugs_report import rapporter
from ..API.sauvegarde import FichierGEO
from .app import white_palette
from .qtlib import GenericThread
from ..pylib.infos import informations_configuration


class Contact(QDialog):
    u"Formulaire utilisé pour contacter l'auteur et rapporter les bogues."

    sent = pyqtSignal(bool, unicode)

    def __init__(self, parent):
        QDialog.__init__(self, parent)
        self.setWindowTitle(u"Contacter l'auteur")
        self.setPalette(white_palette)

        self.parent = parent

        panel = QWidget(self)

        panelSizer = QVBoxLayout()

        avant_propos = QLabel(u"""<i>Afin d'améliorer le fonctionnement de WxGéométrie,
vous êtes invités à signaler tout problème rencontré.</i>""", panel)
        panelSizer.addWidget(avant_propos)
        panelSizer.addSpacing(5)

        rapport = QVBoxLayout()
        rapport_box = QGroupBox(u"Rapport d'incident", panel)
        rapport_box.setLayout(rapport)
        self.titre = titre = QLineEdit(panel)
        titre.setMinimumWidth(200)
        titre.setText(u"Résumé")
        titre.selectAll()
        rapport.addWidget(titre)

        sizer= QHBoxLayout()
        self.modules = modules = QComboBox(panel)
        modules.addItems([parent.onglet(md).titre for md in param.modules if hasattr(self.parent.onglet(md), "titre")])
        modules.setCurrentIndex(self.parent.currentIndex())
        sizer.addWidget(QLabel(u"Module concerné :", panel))
        sizer.addWidget(modules)
        rapport.addLayout(sizer)

        rapport.addWidget(QLabel(u"Description du problème :", panel))
        self.commentaire = commentaire = QTextEdit(panel)
        commentaire.setMinimumSize(200, 150)
        rapport.addWidget(commentaire)

        panelSizer.addWidget(rapport_box)

        sizer = QHBoxLayout()
        coordonnees_box = QGroupBox(u"Vos coordonnées (recommandées)", panel)
        coordonnees_box.setLayout(sizer)
        sizer.addWidget(QLabel(u"Nom :", panel))
        self.nom = nom = QLineEdit(panel)
        nom.setMinimumWidth(100)
        sizer.addWidget(nom)
        sizer.addWidget(QLabel(u" E-mail :"))
        self.mail = mail = QLineEdit(panel)
        mail.setMinimumWidth(100)
        sizer.addWidget(mail)

        panelSizer.addWidget(coordonnees_box)

        btnOK = QPushButton(u"Envoyer", panel)
        btnOK.setToolTip(u"Envoyer les informations.")
        btnCancel = QPushButton(u"Annuler", panel)
        btnCancel.setToolTip(u"Quitter sans rien envoyer.")

        sizer = QHBoxLayout()
        sizer.addWidget(btnOK)
        sizer.addStretch()
        sizer.addWidget(btnCancel)
        panelSizer.addLayout(sizer)

        panel.setLayout(panelSizer)

        topSizer = QHBoxLayout()
        topSizer.addWidget(panel)

        self.setLayout(topSizer)

        self.sent.connect(self.termine)
        btnOK.clicked.connect(self.rapporter)
        btnCancel.clicked.connect(self.close)



    def rapporter(self):
        module = self.parent.onglet(self.modules.currentIndex())

        try:
            histo = module.log.contenu()
        except Exception as e:
            histo = "Impossible de recuperer l'historique du module (%s)." % e

        sys.stdout.flush()
        filename = path2(param.emplacements['log'] + u"/messages.log")
        try:
            file = open(filename, 'r')
            msg = file.read()
        except Exception as e:
            msg = "Impossible de recuperer le journal (%s)." % e
        finally:
            file.close()

        try:
            fgeo = FichierGEO(module=module.nom)
            module._sauvegarder(fgeo)
            fichier_courant = fgeo.data
        except Exception as e:
            fichier_courant = "Impossible de recuperer le fichier courant (%s)." % e

        self.hide()
        kw = {'titre': self.titre.text(), 'auteur': self.nom.text(),
              'email': self.mail.text(), 'description': self.commentaire.toPlainText(),
              'historique': histo, 'log': msg, 'config': informations_configuration(),
              'fichier': fichier_courant
             }
        self.thread = GenericThread(self._rapporter, **kw)
        self.thread.start()
        # XXX: Qt complains here: "QObject::connect: Cannot queue arguments of type 'QTextBlock'"


    def _rapporter(self, **kw):
        # /!\ Ne **JAMAIS** utiliser `print()` depuis une autre thread que la principale !
        self.sent.emit(*rapporter(**kw))



    def termine(self, success, msg):
        if param.debug:
            print(' ** Remote site report **\n' + msg)
        if 'mail()' in msg:
            # Fonction mail() de PHP très probablement bloquée par Free.
            success = False
        if success:
            QMessageBox.information(self, u"Message envoyé", u"Le message a été envoyé avec succès. Merci !")
            self.close()
        else:
            QMessageBox.warning(self, u"Connexion impossible.",
                    u"Impossible d'envoyer le message !\nVous allez être redirigé vers le tracker de bug.")
            self.show()
            # On tente de se connecter directement au tracker.
            # (Le navigateur par défaut est susceptible d'avoir des paramètres proxy correctement renseignés).

            data = urlencode({'item_summary': unicode(self.titre.text()).encode('utf8'),
                            'detailed_desc': unicode(self.commentaire.toPlainText()).encode('utf8'),
                            'anon_email': unicode(self.mail.text()).encode('utf8')})
            open_new_tab("http://wxgeo.free.fr/tracker/?do=newtask&project=1&" + data)




# -*- coding: utf-8 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                   Interpolation                    #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2013  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


from operator import attrgetter

from numpy import array, arange, append

try:
    # wxgeometrie can be imported even if scipy is not found.
    # Only Interpolation_polynomiale_par_morceaux will fail.
    from scipy.interpolate import PiecewisePolynomial
    scipy_found = True
except ImportError:
    scipy_found = False
    def PiecewisePolynomial(*args, **kw):
        raise NotImplementedError, 'Error: scipy library not found !'

from .objet import Ref, Argument, Arguments

from .courbes import Courbe_generique
##from .lignes import Tangente_courbe
from .contexte import contexte

from ..pylib import fullrange
from .. import param



class Interpolation_generique(Courbe_generique):
    u"""Classe mère de toutes les interpolations."""

    points = __points = Arguments("Point_generique")

    _style_defaut = param.interpolations

    def __init__(self, *points, **styles):
        self.__points = points = tuple(Ref(pt) for pt in points)
        Courbe_generique.__init__(self, **styles)


    def _affiche_extremites(self, vec_deb = None, vec_fin = None):
        couleur = self.style("couleur")
        niveau = self.style("niveau")
        epaisseur = self.style("epaisseur")
        debut = self.style("debut")
        fin = self.style("fin")

        # Début de la courbe
        plot = self._representation[-2]
        x, y = self.__points[0].coordonnees
        a, b = self.__points[1].coordonnees
        if vec_deb is None:
            vec_deb = (a - x, b - y)
        if debut is True:
            plot.set_data((x,), (y,))
            plot.set(visible=True, color=couleur, marker="o", markeredgecolor=couleur,
                     markersize=2*self.canvas.taille["o"], linewidth=epaisseur)
            plot.zorder = niveau
        elif debut is False:
            arc = self.rendu.arc(x, y, vec_deb)
            plot.set_data(*arc.get_data()) # À TESTER !!
            plot.set(visible=True, color=couleur, linewidth=epaisseur, marker="")
            plot.zorder = niveau
        else:
            plot.set_visible(False)

        # Fin de la courbe
        plot = self._representation[-1]
        x, y = self.__points[-1].coordonnees
        a, b = self.__points[-2].coordonnees
        if vec_fin is None:
            vec_fin = (a - x, b - y)
        if fin is True:
            plot.set_data((x,), (y,))
            plot.set(visible=True, color=couleur, marker="o", markeredgecolor=couleur,
                     markersize=2*self.canvas.taille["o"], linewidth=epaisseur)
            plot.zorder = niveau
        elif fin is False:
            arc = self.rendu.arc(x, y, vec_fin)
            plot.set_data(*arc.get_data()) # À TESTER !!
            plot.set(visible=True, color=couleur, linewidth=epaisseur, marker="")
            plot.zorder = niveau
        else:
            plot.set_visible(False)




class Interpolation_lineaire(Interpolation_generique):
    u"""Une interpolation linéaire.

    Interpolation entre les points donnés par des segments joints (courbe de classe C0).
    """

    points = __points = Arguments("Point_generique")

    def __init__(self, *points, **styles):
        if styles.get("points", None):
            points = styles.pop("points")
        elif len(points) == 1 and isinstance(points[0], (list, tuple)):
            points = points[0]
        self.__points = points = tuple(Ref(pt) for pt in points)
        Interpolation_generique.__init__(self, *points, **styles)

    def _creer_figure(self):
        n = len(self.__points)
        couleur = self.style("couleur")
        niveau = self.style("niveau")
        style = self.style("style")
        epaisseur = self.style("epaisseur")
        if not self._representation:
            self._representation = [self.rendu.ligne() for i in xrange(n + 1)]

        self._xarray = array([])
        self._yarray = array([])

        if n < 2:
            return

        pas = self.canvas.pas()

        for i in xrange(n - 1):
            plot = self._representation[i]
            x1, y1 = self.__points[i].coordonnees
            x2, y2 = self.__points[i+1].coordonnees
            plot.set_data(array((x1, x2)), array((y1, y2)))
            plot.set(color=couleur, linestyle=style, linewidth=epaisseur)
            plot.zorder = niveau
            # TODO: améliorer l'algo de détection (notamment si l'échelle
            # sur les 2 axes n'est pas la même).
            # Utiliser l'algorithme des segments.
            if abs(y2 - y1) < abs(x2 - x1):
                x_array = arange(x1, x2, pas if x1 < x2 else -pas)
                a = (y2 - y1)/(x2 - x1)
                b = y1 - a*x1
                y_array = a*x_array + b
            elif abs(y2 - y1) > contexte['tolerance']:
                y_array = arange(y1, y2, pas if y1 < y2 else -pas)
                a = (x2 - x1)/(y2 - y1)
                b = x1 - a*y1
                x_array = a*y_array + b
            else:
                continue
            self._xarray = append(self._xarray, x_array)
            self._yarray = append(self._yarray, y_array)

        self._affiche_extremites()


    def image_par(self, transformation):
        # FIXME: l'image n'est plus correcte si le nombre de points est modifié ensuite
        # Par exemple, interp = Interpolation_lineaire(A, B), puis interp.points = (A, B, C)
        return self.__class__(*(transformation(point) for point in self.__points))




class Interpolation_quadratique(Interpolation_generique):
    u"""Une interpolation quadratique.

    Interpolation des points donnés par une courbe polynomiale par morceaux, et de classe C1.
    Pour chaque morceau, x(t)=at²+bt+c et y(t)=dt²+et+f, où t appartient à [0;1].
    """
    points = __points = Arguments("Point_generique")

    def __init__(self, *points, **styles):
        if "points" in styles:
            points = styles.pop("points")
        elif len(points) == 1 and isinstance(points[0], (list, tuple)):
            points = points[0]
        self.__points = points = tuple(Ref(pt) for pt in points)
        Interpolation_generique.__init__(self, *points, **styles)


    def _creer_figure(self):
        n = len(self.__points)
        couleur = self.style("couleur")
        niveau = self.style("niveau")
        style = self.style("style")
        epaisseur = self.style("epaisseur")
        if not self._representation:
            self._representation = [self.rendu.ligne() for i in xrange(n + 1)]

        self._xarray = array([])
        self._yarray = array([])

        if n < 2:
            return

        pas = self.canvas.pas()
        t = fullrange(0, 1, pas)
        for i in xrange(n - 1):
            plot = self._representation[i]
            x0, y0 = self.__points[i].coordonnees
            x1, y1 = self.__points[i+1].coordonnees
            if i == 0:
                dx0, dy0 = x1 - x0, y1 - y0
            a, b, c = x1 - dx0 - x0, dx0, x0
            d, e, f = y1 - dy0 - y0, dy0, y0
            u = (a*t + b)*t + c
            v = (d*t + e)*t + f
            plot.set_data(u, v)
            plot.set(color=couleur, linestyle=style, linewidth=epaisseur)
            plot.zorder = niveau

            dx0 = 2*a + b
            dy0 = 2*d + e

            self._xarray = append(self._xarray, u)
            self._yarray = append(self._yarray, v)

        self._affiche_extremites(vec_fin = (dx0, dy0))




class Interpolation_cubique(Interpolation_generique):
    u"""Une interpolation cubique.

    Interpolation des points donnés par une courbe polynomiale par morceaux, de vecteur tangent horizontal aux sommets, et de classe C1 en général (ie. si x_n!=x_{n-1}).
    Pour chaque morceau, x(t)=at^3+bt^2+ct+d et y(t)=et^3+ft^2+gt+h, où t appartient à [0;1], et y'(0)=y'(1)=0.
    """

    points = __points = Arguments("Point_generique")

    def __init__(self, *points, **styles):
        if "points" in styles:
            points = styles.pop("points")
        elif len(points) == 1 and isinstance(points[0], (list, tuple)):
            points = points[0]
        self.__points = points = tuple(Ref(pt) for pt in points)
        Interpolation_generique.__init__(self, *points, **styles)


    def _creer_figure(self):
        n = len(self.__points)
        couleur = self.style("couleur")
        niveau = self.style("niveau")
        style = self.style("style")
        epaisseur = self.style("epaisseur")
        courbure = self.style("courbure")
        if courbure is None:
            courbure = 1
        if not self._representation:
            self._representation = [self.rendu.ligne() for i in xrange(n + 1)]

        self._xarray = array([])
        self._yarray = array([])

        if n < 2:
            return

        pas = self.canvas.pas()
        t = fullrange(0, 1, pas)
        for i in xrange(n - 1):
            plot = self._representation[i]
            x0, y0 = self.__points[i].coordonnees
            x1, y1 = self.__points[i+1].coordonnees
            if i == 0:
                dx0 = x1 - x0
                dy0 = y1 - y0
            else:
                dx0 = courbure*abs(x1 - x0)
                if dx1 != 0:
                    dy0 = dy1/dx1*dx0
                else:
                    dx0 = dx1
                    dy0 = dy1
            if i < n - 2:
                x2, y2 = self.__points[i+2].coordonnees
                if cmp(y0, y1) == cmp(y1, y2):
                    dy1 = y2 - y0
                    dx1 = x2 - x0
                else:
                    dy1 = 0
                    dx1 = courbure*abs(x1 - x0)
            else:
                dy1 = y1 - y0
                dx1 = x1 - x0
            a = 2*(x0 - x1) + dx0 + dx1
            b = 3*(x1 - x0) -2*dx0 - dx1
            c = dx0
            d = x0
            e = 2*(y0 - y1) + dy0 + dy1
            f = 3*(y1 - y0) -2*dy0 - dy1
            g = dy0
            h = y0
            u = ((a*t + b)*t + c)*t + d
            v = ((e*t + f)*t + g)*t + h
            plot.set_data(u, v)
            plot.set(color=couleur, linestyle=style, linewidth=epaisseur)
            plot.zorder = niveau

            self._xarray = append(self._xarray, u)
            self._yarray = append(self._yarray, v)

        self._affiche_extremites(vec_deb = (dx0, dy0), vec_fin = (dx1, dy1))


class Interpolation_polynomiale_par_morceaux(Interpolation_generique):
    u"""Une courbe d'interpolation polynomiale par morceaux.
    Elle utilise l'interpolation par morceau de scipy pour construire la fonction
    c'est la classe scipy.interpolate.PiecewisePolynomial

    elle passe par les points (xl, yl) et avec le nombre derive dans derivl à
    l'abscisse xl.

    exemple::

    >>> from wxgeometrie import *
    >>> A = Point(-1,-2)
    >>> B = Point(2,1)
    >>> C = Point(8,-3)
    >>> d = Interpolation_polynomiale_par_morceaux(A,B,C, derivees=[-1,0.5,2])

    :type derivee: list
    :param derivee: value of the diff number at each point

    :rtype : numpy.lib.polynomial

    """
    _style_defaut = param.interpolations_polynomiales_par_morceaux

    points = __points = Arguments("Point_interpolation")
    fonction = None # fonction d'interpolation

    def __init__(self, *points, **styles):
        debut = styles.pop("debut", True)
        fin = styles.pop("fin", True)
        if styles.get("points", None):
            points = styles.pop("points")
        # Accepter Interpolation_polynomiale_par_morceaux([A, B, C])
        # comme alias de Interpolation_polynomiale_par_morceaux(A, B, C).
        elif len(points) == 1 and isinstance(points[0], (list, tuple)):
            points = points[0]
        self.__points = points = tuple(Ref(pt) for pt in points)
        # en test: creation des tangentes: dictionnaire
        # key: nom du point, value: objet Tangente_courbe
        ##self.__tangentes = []
        # boucle avec ruse enumerate
        # cf http://docs.python.org/tutorial/datastructures.html#dictionaries
        ##for i, P in enumerate(points):
            ##dico = {'point': P, 'cdir': self.__derivees[i]}
            ##self.__tangentes.append(Tangente_courbe(**dico))

        Interpolation_generique.__init__(self, *points, **styles)


    @property
    def fonction(self):
        """Fonction wrapper vers la fonction de scipy PiecewisePolynomial

        """
        pts = self.points_tries
        xl = [P.x for P in pts]
        yl = [P.y for P in pts]
        yl_cum = zip(yl, self._derivees())
        return PiecewisePolynomial(xl, yl_cum)


    def _derivees(self):
        derivees = []
        points = self.points_tries
        n = len(points)
        for i, P in enumerate(points):
            if P.derivee is None:
                if n == 1:
                    derivees.append(0)
                elif i == 0:
                    # Premier point de la courbe
                    A, B = points[:2]
                    dy = B.y - A.y
                    dx = B.x - A.x
                    derivees.append(dy/dx if dx else 0)
                elif i == n - 1:
                    # Dernier point de la courbe
                    A, B = points[-2:]
                    dy = B.y - A.y
                    dx = B.x - A.x
                    derivees.append(dy/dx if dx else 0)
                else:
                    # Il y a (au moins) un point avant et un point après
                    A, B, C = points[i - 1:i + 2]
                    if B.y >= max(A.y, C.y) or B.y <= min(A.y, C.y):
                        derivees.append(0)
                    else:
                        der = self.strategies.get(self.style('strategie'),
                                            self._strategie_pente_moyenne)
                        derivees.append(der(self, A, B, C))
            else:
                derivees.append(P.derivee)
        return derivees


    def _strategie_pente_moyenne(self, A, B, C):
        u"""La pente en B est celle de la droite (AC).

        Stratégie simple et qui donne un rendu assez esthétique en pratique.
        """
        dy = C.y - A.y
        dx = C.x - A.x
        return dy/dx if dx else 0

    def _strategie_pente_minimale(self, A, B, C):
        u"""Pente la plus faible entre celle de (AB) et celle de (BC).

        On prend la pente la plus faible, de façon à être sûr
        qu'à gauche et à droite du point considéré, la courbe d'interpolation
        reste comprise en ordonnées entre le point considéré et
        le point suivant.
        Cela facilite la construction d'extrema :
        si A, B, C sont trois points d'interpolation,
        avec B.y < A.y et B.y < C.y, alors on est assuré que
        la courbe ne descendra pas en dessous de B.y sur
        l'intervalle [A.x, C.x].
        """
        dy1 = B.y - A.y
        dx1 = B.x - A.x
        der1 = (dy1/dx1 if dx1 else 0)
        dy2 = C.y - B.y
        dx2 = C.x - B.x
        der2 = (dy2/dx2 if dx2 else 0)
        return der1 if abs(der1) < abs(der2) else der2

    def _strategie_moyenne_gauche_droite(self, A, B, C):
        u"""
        Moyenne de la pente moyenne à gauche et de la pente moyenne à droite.

        Stratégie peu concluante en pratique.
        Il faudrait utiliser un barycentre plutôt qu'une moyenne simple.
        """
        dy1 = B.y - A.y
        dx1 = B.x - A.x
        der1 = (dy1/dx1 if dx1 else 0)
        dy2 = C.y - B.y
        dx2 = C.x - B.x
        der2 = (dy2/dx2 if dx2 else 0)
        return .5*(der1 + der2)


    strategies = {'pente_moyenne': _strategie_pente_moyenne,
                  'pente_minimale': _strategie_pente_minimale,
                  'moyenne_gauche_droite': _strategie_moyenne_gauche_droite
                  }


    @property
    def points_tries(self):
        # TODO: mise en cache
        return sorted(self.__points, key=attrgetter('x'))


    def _creer_figure(self):
        points = self.points_tries
        n = len(points)
        couleur = self.style("couleur")
        niveau = self.style("niveau")
        style = self.style("style")
        epaisseur = self.style("epaisseur")
        if not self._representation:
            self._representation = [self.rendu.ligne() for i in xrange(3)]

        if n < 2:
            return
        # de même les tangentes sont recalculées
        # il doit y avoir moyen de faire moins de calculs
        # self.__tangentes = []
        # for i in range(len(self.__points)):
        #     self.__tangentes.append(Tangente_courbe(point = self.__points[i],\
        #                                                 cdir = self.__derivees[i]))
        #     #self.__tangentes[i]._creer_figure()

        pas = self.canvas.pas()
        plot = self._representation[0]
        x1, y1 = points[0].coordonnees
        x2, y2 = points[-1].coordonnees
        xarray = fullrange(x1, x2, pas)
        yarray = self.fonction(xarray)
        plot.set_data(xarray, yarray)
        plot.set(color=couleur, linestyle=style, linewidth=epaisseur)
        plot.zorder = niveau
        self._xarray = xarray
        self._yarray = yarray

        self._affiche_extremites()

    @property
    def xmin(self):
        return self.points_tries[0].x if self.__points else None

    @property
    def xmax(self):
        return self.points_tries[-1].x if self.__points else None




# -*- coding: utf-8 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

from wxgeometrie.mathlib.custom_objects import Decim
from sympy import Symbol, Rational

from tools.testlib import assertEqual

x = Symbol('x')


def test_Decim():
    assertEqual(repr(Decim(1, 2)*x + Decim(1, 5)), '0.5*x + 0.2')
    assertEqual(repr(Decim(1, 2)*Rational(1, 5)), '0.1')


# -*- coding: utf-8 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)
from __future__ import with_statement

##--------------------------------------##
#                  Onglets               #
##--------------------------------------##
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2013  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
from webbrowser import open_new_tab
from functools import partial
from cStringIO import StringIO

from PyQt4.QtGui import QTabWidget, QToolButton, QIcon, QMessageBox, QFileDialog, \
                        QDialog, QPainter, QPrintDialog, QPrinter, QFont
from PyQt4.QtCore import Qt, QPoint, QByteArray, QRectF
from PyQt4.QtSvg import QSvgRenderer
import matplotlib.backend_bases as backend_bases

from .aide import About, Informations
from .animer import DialogueAnimation
from .contact import Contact
from .dialogues_geometrie import EditerObjet, SupprimerObjet
from .fenetre_options import FenetreOptions
from .inspecteur import FenCode
from .nouvelles_versions import Gestionnaire_mises_a_jour
from .proprietes_feuille import ProprietesFeuille
from .proprietes_objets import Proprietes
from .qtlib import PopUpMenu, png_pth
from . import dialogues_geometrie
from ..API.sauvegarde import FichierGEO, ouvrir_fichierGEO
from .. import param, modules, geolib
from ..param import NOMPROG
from ..pylib import print_error, debug, path2
from ..param.options import options as param_options


class Onglets(QTabWidget):

    _nom_fichier_session = ''

    def __init__(self, parent):
        self.parent = parent
        QTabWidget.__init__(self, parent)
        self.setStyleSheet("""
            QTabBar::tab:selected {
            background: white;
            border: 1px solid #C4C4C3;
            border-bottom-color: white; /* same as the pane color */
            border-top-left-radius: 4px;
            border-top-right-radius: 4px;
            min-width: 8ex;
            padding: 7px;
            }
            QStackedWidget {background:white}
            QTabBar QToolButton {
            background:white;
            border: 1px solid #C4C4C3;
            border-bottom-color: white; /* same as the pane color */
            border-top-left-radius: 4px;
            border-top-right-radius: 4px;
            }
            """)
        self.setTabsClosable(False)
        self.setMovable(True)
        ##self.tabCloseRequested.connect(self.fermer_onglet)
#        palette = QPalette()
#        white = QColor(Qt.white)
#        palette.setColor(QPalette.Window, white)
#        palette.setColor(QPalette.Button, white)
#        palette.setColor(QPalette.WindowText, white)
#        self.setPalette(palette)

        ###############################
        # Creation de fonctions associees aux entrees du menu "Creer"
        self.creer = {}
        DG = dialogues_geometrie.__dict__
        dialogues = [(nom[8:], DG[nom]) for nom in DG if nom.startswith("Dialogue")]
        for dialogue in dialogues:
            def f(event = None, self = self, dialogue = dialogue[1]):
                self.creer_objet(dialogue)
            self.creer[dialogue[0]] = f
        ###############################

        # Bouton "Nouvel onglet"
        self.newTabButton = newTabButton = QToolButton(self)
        self.setCornerWidget(newTabButton, Qt.TopLeftCorner)
        newTabButton.setCursor(Qt.ArrowCursor)
        newTabButton.setAutoRaise(True)
        newTabButton.setIcon(QIcon(path2("%/wxgeometrie/images/newtab3.png")))
        newTabButton.clicked.connect(self.popup_activer_module)
        newTabButton.setToolTip(u"Activer un autre onglet")

        # Bouton "Fermer l'onglet"
        self.closeTabButton = closeTabButton = QToolButton(self)
        self.setCornerWidget(closeTabButton, Qt.TopRightCorner)
        closeTabButton.setCursor(Qt.ArrowCursor)
        closeTabButton.setAutoRaise(True)
        closeTabButton.setIcon(QIcon(path2("%/wxgeometrie/images/closetab.png")))
        closeTabButton.clicked.connect(partial(self.fermer_onglet, None))
        closeTabButton.setToolTip(u"Fermer l'onglet courant")

        self.gestionnaire_de_mises_a_jour = Gestionnaire_mises_a_jour(self)

        self._liste = [] # liste des onglets

        # Ajoute les differentes composantes :
        self.actualiser_liste_onglets()

        # Usage interne : le signal `.currentChanged` ne transmet pas le numéro de l'onglet
        # précédemment actif, il faut donc garder manuellement en mémoire l'onglet actif.
        self._ancien_onglet = None
        # adaptation du titre de l'application et du menu.
        self.currentChanged.connect(self.evt_changer)

        # Use the tabMoved signal of QTabBar. You can get the QTabBar used
        # in QTabWidget with the tabBar() method.
        self.tabBar().tabMoved.connect(self.evt_deplacer_onglet)


    # -------------------
    # Gestion des onglets
    # -------------------

    def popup_activer_module(self):
        u"""Affiche un menu permettant d'activer des modules."""
        menu = PopUpMenu(u"Module à activer", self, 'crayon')
        deja_charges = [onglet.__module__.split('.')[-1] for onglet in self]
        exercices = None
        for nom in param.modules:
            if nom not in deja_charges:
                titre = param.descriptions_modules[nom]['titre']
                if titre.startswith('Exercices - '):
                    if exercices is None:
                        exercices = menu.addMenu(u'Exercices')
                    action = exercices.addAction(titre[12:])
                else:
                    action = menu.addAction(titre)
                action.triggered.connect(partial(self.activer_module, nom, selectionner=True))
        menu.addSeparator()
        action = menu.addAction(QIcon(png_pth('reload')), u"Restaurer la session précédente")
        action.setIconVisibleInMenu(True)
        action.triggered.connect(self.ChargerSessionPrecedente)
        font = QFont()
        ##font.setBold(True)
        font.setStyle(QFont.StyleItalic)
        action.setFont(font)
        menu.exec_(self.newTabButton.mapToGlobal(QPoint(0, self.newTabButton.height())))

    def activer_module(self, nom, selectionner=True):
        u"""Active le module `nom`.

        Retourne `True` si le module a bien été activé (ou est déjà actif),
        `False` sinon."""
        if nom not in param.modules_actifs:
            print(u"Warning: Le module %s n'a pas été trouvé." %repr(nom))
            return False
        elif param.modules_actifs[nom]:
            print(u'Le module %s est déjà activé.' %repr(nom))
        else:
            param.modules_actifs[nom] = True
            module = modules.importer_module(nom)
            if module is None:
                print(u"Warning: Impossible d'importer le module %s." %nom)
                return False
            panel = module._panel_(self, module)
            self.nouvel_onglet(panel)
            if selectionner:
                self.changer_onglet(panel)
        return True

    def nouvel_onglet(self, panel, i=None):
        u"Ajouter un nouvel onglet à la position 'i'."
        if i is None:
            self._liste.append(panel)
            self.addTab(panel, panel.titre)
        else:
            self._liste.insert(i, panel)
            self.insertTab(i, panel, panel.titre)
        setattr(self, panel.module._nom_, panel)
        if self.count() > 1:
            self.closeTabButton.setEnabled(True)

    def deplacer_onglet(self, i, j):
        u"Déplace l'onglet situé en position `i` à la position `j`."
        if i != j:
            self.tabBar.moveTab(i, j)

    def fermer_onglet(self, i=None):
        u"Ferme l'onglet situé en position `i`."
        if self.count() > 1:
            if i is None:
                i = self.currentIndex()
            panel = self._liste.pop(i)
            delattr(self, panel.module._nom_)
            self.deleteTab(i)
            param.modules_actifs[panel.nom] = False
        if self.count() <= 1:
            self.closeTabButton.setEnabled(False)

    def deleteTab(self, i):
        u"""Supprime le ième onglet.

        Ne pas utiliser directement ; utiliser `fermer_onglet()` à la place."""
        tab = self.widget(i)
        self.removeTab(i)
        tab.close()
        tab.deleteLater()

    def evt_changer(self, index):
        u"""Actions effectuées lorsqu'on change d'onglet."""
        if self._ancien_onglet is not None:
            self._ancien_onglet.desactiver()
        if index != -1:
            onglet = self._liste[index]
            self.actualise_onglet(onglet)
            # Le module nouvellement sélectionné peut exécuter
            # des actions personnalisées à cette occasion.
            onglet.activer()
            self._ancien_onglet = onglet
        else:
            self._ancien_onglet = None


    def evt_deplacer_onglet(self, new_index, old_index):
        # Mise à jour de self._liste
        # XXX: self._liste est-il toujours utile depuis le passage à Qt ?
        self._liste.insert(new_index, self._liste.pop(old_index))


    def actualise_onglet(self, onglet):
        self.parent.setMenuBar(onglet.module._menu_(onglet)) # change le menu de la fenetre
        onglet.changer_titre() # change le titre de la fenetre


    def onglet(self, nom):
        u"nom : nom ou numéro de l'onglet."
        if type(nom) == int:
            return self._liste[nom]
        return getattr(self, nom, None)


    @property
    def onglet_actuel(self):
        if self._liste:
            return self._liste[self.currentIndex()]

    def onglet_suivant(self, event):
        self.setCurrentIndex((self.currentIndex() + 1) % self.count())

    def __nonzero__(self):
        return bool(self._liste)

    def __iter__(self):
        return iter(self._liste)

    def changer_onglet(self, onglet):
        u"""Changer d'onglet actif.

        `onglet` : l'onglet proprement dit, ou son nom, ou son numéro."""
        if type(onglet) in (str, unicode):
            onglet = self._liste.index(getattr(self, onglet.lower()))
        elif type(onglet) not in (int, long):
            onglet = self._liste.index(onglet)
        self.setCurrentIndex(onglet)


    def actualiser_liste_onglets(self):
        # On commence par vérifier que la liste des modules à charger n'est pas vide
        modules_a_charger = [nom for nom, val in param.modules_actifs.iteritems() if val]
        print u"Modules à charger :", modules_a_charger
        if not modules_a_charger:
            # On affiche l'onglet de bienvenue par défaut.
            modules_a_charger = ['bienvenue']

        # `pos` indique la progression du classement:
        # tous les onglets situés avant `pos` sont déjà classés.
        pos = 0
        for nom in modules_a_charger:
            try:
                module = modules.importer_module(nom)
                if module is not None:
                    for i, panel in enumerate(self._liste[pos:]):
                        if panel.module is module:
                            # Déplacer le panel en position pos
                            self.deplacer_onglet(pos + i, pos)
                            # Mettre à jour la position
                            pos += 1
                            break
                    else:
                        # Créer un onglet en position pos
                        self.nouvel_onglet(module._panel_(self, module), pos)
                        # Mettre à jour la position
                        pos += 1
            except Exception:
                param.modules_actifs[nom] = False
                print("\n** MODULE '%s' : ERREUR FATALE **\n" % nom)
                print_error()
                msg = ("Le programme va redémarrer sans le module '%s'." % nom)
                print('\n' + len(msg)*'-')
                print(msg)
                print("Veuillez patienter...")
                print(len(msg)*'-' + '\n')
                # Ne pas redémarrer pendant l'initialisation... (appeler la
                # méthode .close() à l'intérieur de la méthode .__init__()
                # ne fonctionne pas).
                param._restart = True

        # Supprimer tous les onglets qui sont situés après pos
        while pos < self.count():
            self.fermer_onglet(pos)


    # -------------------
    # Sauvegardes, export
    # -------------------

    filtres_save = (u"Fichiers " + NOMPROG + u" (*.geo);;"
                    u"Fichiers " + NOMPROG + u" compressés (*.geoz);;"
                    u"Tous les fichiers (*.*)")

    filtre_session = u"Session Géophar (*.geos)"

    def NewFile(self):
        self.onglet_actuel.creer_feuille()


    def SaveFile(self):
        if self.onglet_actuel.nom_sauvegarde:
            self.onglet_actuel.sauvegarder()
        else:
            # Pas de nom par défaut.
            self.SaveFileAs()


    def SaveFileAs(self):
        actuelle = self.onglet_actuel.feuille_actuelle # feuille de travail courante
        if actuelle and actuelle.sauvegarde["nom"]:
            fichier = actuelle.sauvegarde["nom"] # le nom par defaut est le nom précédent
            dir = actuelle.sauvegarde["repertoire"]
        else:
            fichier = ""
            if param.rep_save is None:
                dir = param.repertoire
            else:
                dir = param.rep_save
        filtre = (u"Fichiers %s compressés (*.geoz)" if param.compresser_geo else u'')
        path, filtre = QFileDialog.getSaveFileNameAndFilter(self, u"Enregistrer sous ...",
                                   os.path.join(dir, fichier), self.filtres_save, filtre)

        if path:
            # Sauvegarde le répertoire pour la prochaine fois
            param.rep_save = os.path.dirname(path)
            if param.rep_open is None:
                param.rep_open = param.rep_save
            if param.rep_export is None:
                param.rep_export = param.rep_save

            self.onglet_actuel.sauvegarder(path)


    def OpenFile(self, _event=None, detecter_module=True):
        # `_event=None` est un argument factice, qui est sert à consommer le booléen
        # renvoyé en argument lorsque OpenFile est lié au signal QAction.triggered.
        # Cf. doc Qt :
        # void QAction::triggered ( bool checked = false )   [signal]
        if param.rep_open is None:
            dir = param.repertoire
        else:
            dir = param.rep_open
        filtre = (u"Fichiers %s compressés (*.geoz)" if param.compresser_geo else u'')
        paths, filtre = QFileDialog.getOpenFileNamesAndFilter(self, u"Choisissez un fichier", dir,
                                             self.filtres_save, filtre)
        if paths:
            # Sauvegarde le répertoire pour la prochaine fois
            param.rep_open = os.path.dirname(paths[0])
            if param.rep_save is None:
                param.rep_save = param.rep_open
            if param.rep_export is None:
                param.rep_export = param.rep_open

            for path in paths:
                if detecter_module: # on detecte dans quel onglet le fichier doit etre ouvert
                    self.ouvrir(path)
                else: # on l'ouvre dans l'onglet actuellement ouvert
                    self.onglet_actuel.ouvrir(path) # ouvre tous les fichiers selectionnes



    def OpenFileHere(self):
        u"""Ouvrir le fichier dans le module courant.

        Par défaut, sinon, le fichier est ouvert dans le module qui l'a crée."""
        self.OpenFile(detecter_module=False)


    def ouvrir(self, fichier, en_arriere_plan = False):
        u"""Ouvre un fichier dans l'onglet adéquat.

        'fichier' est soit l'adresse d'un fichier .geo, soit une instance de FichierGEO.
        """
        if not isinstance(fichier, FichierGEO):
            fichier, message = ouvrir_fichierGEO(fichier)
            self.parent.message(message)
        module = self.onglet(fichier.module)
        if module is None:
            if self.activer_module(fichier.module, selectionner=(not en_arriere_plan)):
                module = self.onglet(fichier.module)
            else:
                self.parent.message(u"Le module '%s' n'a pas été trouvé." % fichier.module)
                return
        if not en_arriere_plan:
            self.changer_onglet(module) # affiche cet onglet
        module.ouvrir(fichier) # charge le fichier dans le bon onglet


    def ExportFile(self, lieu = None, sauvegarde = False, exporter = True):
        u"""Le paramètre sauvegarde indique qu'il faut faire une sauvegarde simultanée.
        (attention, on ne vérifie pas que le fichier .geo n'existe pas !).
        """

        actuelle = self.onglet_actuel.feuille_actuelle # feuille de travail courante
        if actuelle and actuelle.sauvegarde["export"]:
            dir, fichier = os.path.split(actuelle.sauvegarde["export"]) # on exporte sous le même nom qu'avant par défaut
        elif actuelle and actuelle.sauvegarde["nom"]:
            fichier = actuelle.sauvegarde["nom"] # le nom par defaut est le nom de sauvegarde
            dir = actuelle.sauvegarde["repertoire"]
        else:
            fichier = ""
            if param.rep_export is None:
                dir = param.repertoire
            else:
                dir = param.rep_export


        lieu = lieu or self.onglet_actuel.nom # par defaut, le lieu est l'onglet courant

        description_formats = sorted(backend_bases.FigureCanvasBase.filetypes.iteritems())
        formats_supportes = sorted(backend_bases.FigureCanvasBase.filetypes)


        def format(typ, description):
            return typ.upper() + ' - ' + description + ' (*.' + typ + ')'
        filtres = [format(k, v) for k, v in description_formats]

        format = ''
        if '.' in fichier and fichier[-1] != '.':
            # On récupère l'extension du nom de fichier
            format = fichier.rsplit('.', 1)[1]

        if format not in formats_supportes:
            format = param.format_par_defaut
            if format not in formats_supportes:
                print("Warning: format par defaut incorrect: " + repr(format))
                format = 'png'

        for filtre in filtres:
            if filtre.endswith(format + ')'):
                break

        path, filtre = QFileDialog.getSaveFileNameAndFilter(self, u"Exporter l'image",
                                           os.path.join(dir, fichier), ';;'.join(filtres),
                                           filtre)

        if not path:
            return # Quitte sans sauvegarder.

        # Vérifie si le nom de fichier possède une extension
        for format in formats_supportes:
            if path.lower().endswith(format):
                break
        else:
            # Sinon, on en rajoute une.
            i = filtre.rindex('.')
            format = filtre[i + 1:-1]
            path += '.' + format

        # Sauvegarde le répertoire pour la prochaine fois
        param.rep_export = os.path.dirname(path)
        if param.rep_save is None:
            param.rep_save = param.rep_export
        if param.rep_open is None:
            param.rep_open = param.rep_export
        if sauvegarde:
            param.rep_save = param.rep_export
        # Sauvegarde le format pour la prochaine fois
        param.format_par_defaut = format
        ##print 'param.format_par_defaut', param.format_par_defaut

        try:
            if sauvegarde:
                geofile_name = path.rsplit('.', 1)[0] + ('.geoz' if param.compresser_geo else '.geo')
                self.onglet(lieu).sauvegarder(geofile_name)
            # Save Bitmap
            if exporter:
                self.onglet(lieu).exporter(path)
        except:
            print_error()
            self.parent.message("Erreur lors de l'export.")
        return path


    def ExportAndSaveFile(self, lieu=None):
        self.ExportFile(lieu=lieu, sauvegarde=True)

    def CloseFile(self):
        self.onglet_actuel.fermer_feuille()

    def NouvelleSession(self):
        reponse = QMessageBox.question(self, u'Ouvrir une session vierge ?',
                                       u'Voulez-vous démarrer une nouvelle session ?\n'
                                       u'Attention, la session actuelle sera perdue.',
                                       QMessageBox.Yes | QMessageBox.No,
                                       QMessageBox.Yes)
        if reponse == QMessageBox.Yes:
            self.parent.gestion.reinitialiser_session()

    def SauverSession(self):
        filtre = self.filtre_session
        path = self._nom_fichier_session
        path, filtre = QFileDialog.getSaveFileNameAndFilter(self, u"Sauver la session sous ...",
                                           path, filtre, filtre)
        if not path.endswith('.geos'):
            path += '.geos'
        if path:
            # Mémorise l'emplacement pour la prochaine fois
            self._nom_fichier_session = path
            # Enregistrement de la session
            self.parent.gestion.sauver_session(lieu=path, force=True)

    def ChargerSession(self):
        filtre = self.filtre_session
        path = self._nom_fichier_session
        path, filtre = QFileDialog.getOpenFileNameAndFilter(self,
                            u"Choisissez un fichier de session",
                            path or path2(param.emplacements['session']),
                            filtre, filtre)
        if path:
            if not path.endswith('.geos'):
                path += '.geos'
            # Mémorise l'emplacement pour la prochaine fois
            self._nom_fichier_session = path
            # Charge la session
            self.parent.gestion.charger_session(lieu=path)

    def ChargerSessionPrecedente(self):
        reponse = QMessageBox.question(self, u'Recharger la session précédente ?',
                                       u'Voulez-vous recharger la session précédente ?\n'
                                       u'Attention, la session actuelle sera perdue.',
                                       QMessageBox.Yes | QMessageBox.No,
                                       QMessageBox.Yes)
        if reponse == QMessageBox.Yes:
            self.parent.gestion.charger_session()

    # ----------
    # Impression
    # ----------

    ##def PageSetup(self):
        ##self.a_venir()

    def Printout(self):
        # Version utilisant QSvg
        # Fixme: l'alpha n'est pas bien rendu
        # (le fond des polygones est absent)
        # Le fond est bien présent après export avec matplotlib pourtant.
        printer = QPrinter(QPrinter.HighResolution)
        dialog = QPrintDialog(printer, self)
        dialog.setOption(QPrintDialog.PrintPageRange, False)
        dialog.setOption(QPrintDialog.PrintToFile, True)
        dialog.setOption(QPrintDialog.PrintShowPageSize, True)
        dialog.setWindowTitle("Imprimer le document")
        if (dialog.exec_() == QDialog.Accepted):
            painter = QPainter(printer)
            try:
                output = StringIO()
                self.onglet_actuel.canvas.exporter(output, format='svg')
                svg = QSvgRenderer(QByteArray(output.getvalue()), self)
                size = svg.defaultSize()
                size.scale(painter.viewport().size(), Qt.KeepAspectRatio)
                rect = QRectF(0, 0, size.width(), size.height())
                svg.render(painter, rect)
            finally:
                painter.end()


    ##def Printout(self):
        ##printer = QPrinter(QPrinter.HighResolution)
        ##dialog = QPrintDialog(printer, self)
        ##dialog.setOption(QPrintDialog.PrintPageRange, False)
        ##dialog.setOption(QPrintDialog.PrintToFile, True)
        ##dialog.setOption(QPrintDialog.PrintShowPageSize, True)
        ##dialog.setWindowTitle("Imprimer le document")
        ##if (dialog.exec_() == QDialog.Accepted):
            ##painter = QPainter(printer)
            ##psize = printer.pageRect(QPrinter.Inch)
            ### QSizeF -> tuple
            ##tsize = float(psize.width())*2.54, float(psize.height())*2.54
            ##print printer.resolution(), tsize
            ##img = self.onglet_actuel.canvas.as_QImage(dpi=printer.resolution(),
                                                      ##taille=tsize,
                                                      ##keep_ratio=True)
            ##size = img.size()
            ##rect = painter.viewport()
            ##size.scale(rect.size(), Qt.KeepAspectRatio)
            ##painter.setViewport(rect.x(), rect.y(), size.width(), size.height())
            ##painter.setWindow(img.rect())
            ##painter.drawImage(0, 0, img)
            ##painter.end()

    def a_venir(self):
        QMessageBox.information(self, u"A venir !", u"Fonctionnalité non présente pour l'instant !")


    # ---------------------------------------------
    # Boîtes de dialogue liées à la feuille ouverte
    # ---------------------------------------------

    def supprimer(self):
        canvas = self.onglet_actuel.canvas
        dlg = SupprimerObjet(self)
        if dlg.exec_() == QDialog.Accepted:
            with canvas.geler_affichage(actualiser=True, sablier=True):
                for selection in dlg.selectedItems():
                    try:
                        # Il est normal que des erreurs soient renvoyées
                        # si un objet dépendant d'un autre est déjà supprimé.
                        objet = selection.text().split()[1]
                        canvas.feuille_actuelle.objets[objet].supprimer()
                    except Exception:
                        print_error()
                canvas.feuille_actuelle.interprete.commande_executee()


    def editer(self):
        feuille = self.onglet_actuel.feuille_actuelle
        if feuille:
            objets = []
            dlg = EditerObjet(self)
            if dlg.exec_() == QDialog.Accepted:
                objets = [feuille.objets[selection.text().split()[1]] for selection in dlg.selectedItems()]

            if objets:
                win = Proprietes(self, objets)
                win.show()


    def creer_objet(self, classe):
        canvas = self.onglet_actuel.canvas
        dl = classe(self)
        while dl.affiche() == QDialog.Accepted:
            try:
                canvas.executer(dl.commande(), parser = True)
                break
            except NameError:
                print_error()
                canvas.message(u'Erreur : nom déjà utilisé ou nom réservé.')
            except:
                print_error()
                canvas.message(u'Erreur : paramètres incorrects.')


    def Animer(self):
        d = DialogueAnimation(self)
        d.show()

    def Histo(self):
        contenu = self.onglet_actuel.feuille_actuelle.sauvegarder()
        h = FenCode(self, u"Contenu interne de la feuille", contenu, self.executer_dans_feuille_courante)
        h.show()

    def executer_dans_feuille_courante(self, instructions):
        self.onglet_actuel.creer_feuille()
        self.onglet_actuel.feuille_actuelle.charger(instructions)

    def Proprietes(self):
        actuelle = self.onglet_actuel.feuille_actuelle # feuille courante
        ProprietesFeuille(self, actuelle).show()


    # -------------------------
    # Autres boîtes de dialogue
    # -------------------------

    def Options(self):
        fen_options = FenetreOptions(self, param_options)
        fen_options.options_modified.connect(self.apply_options)
        fen_options.show()

    def apply_options(self):
        self.actualiser_liste_onglets()
        for parametre in ('decimales', 'unite_angle', 'tolerance'):
            geolib.contexte[parametre] = getattr(param, parametre)

    def Aide(self):
        open_new_tab(path2("%/wxgeometrie/doc/html/help.html"))

    def Contacter(self):
        self.formulaire = Contact(self)
        self.formulaire.show()

    def About(self):
        dialog = About(self)
        dialog.exec_()

    def Informations(self):
        dialog = Informations(self)
        dialog.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                Calculatrice                 #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2013  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from functools import partial

from PyQt4.QtGui import (QHBoxLayout, QVBoxLayout, QCheckBox, QIcon, QPushButton,
                         QTextEdit, QMenu, QLabel, QSpinBox, QCursor, QTextCursor,
                         QToolButton, QWidget, QTabWidget, QGroupBox, QComboBox,
                         QLineEdit)
from PyQt4.QtCore import Qt

from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

from ...GUI.ligne_commande import LigneCommande
from ...GUI.qtlib import png
from ...GUI.inspecteur import FenCode
from ...GUI.menu import MenuBar
from ...GUI.panel import Panel_simple
from ...mathlib.interprete import Interprete
from ...mathlib.parsers import latex2mathtext
from ...mathlib.end_user_functions import __classement__

from ...pylib import print_error, uu, debug, no_argument, eval_safe
from ... import param


class CalculatriceMenuBar(MenuBar):
    def __init__(self, panel):
        MenuBar.__init__(self, panel)
        self.ajouter(u"Fichier", [u"Réinitialiser",
                                    u"Réinitialiser la calculatrice.", "Ctrl+N",
                                    self.panel.initialiser],
                                [u"ouvrir"], [u"enregistrer"],
                                [u"enregistrer_sous"], ['session'], None, ["quitter"])
        self.ajouter(u"Affichage", ["onglet"], ["plein_ecran"])
        for rubrique in __classement__:
            self.ajouter(rubrique, *(self.formater(contenu, rubrique != "Symboles") for contenu in __classement__[rubrique]))
            # pas de parenthese apres un symbole
        self.ajouter(u"Outils",
                        [u"Mémoriser le résultat", u"Copie le resultat du calcul dans le presse-papier, afin de pouvoir l'utiliser ailleurs.", "Ctrl+M", self.panel.vers_presse_papier],
                        [u"options"],
                        )
        self.ajouter(u"Avancé",
                        [u"État interne de l'interprète", u"État de l'interprète de commandes.", u"Ctrl+H", self.panel.EtatInterne],
                        [u"ligne_commande"],
                        ["debug"],
                        )
        self.ajouter("?")


    def formater(self, contenu, parentheses = True):
        if contenu is None:
            return
        titre, nom, doc = contenu
        return [titre, doc, "", partial(self.panel.insere, nom=nom, parentheses=parentheses)]



class BoutonValider(QToolButton):

    modes = [('exact', u'résultats exacts'), ('approche', u'résultats approchés'),
             ('scientifique', u'résultats en écriture scientifique')]

    def __init__(self, parent):
        QToolButton.__init__(self)
        self.parent = parent
        self.setAutoRaise(True)
        self.mode_normal()

        self.menu = menu = QMenu(self)
        self.setMenu(menu)
        for mode, titre in self.modes:
            nom = 'mode_' + mode
            action = menu.addAction(QIcon(png(nom)), titre)
            action.setIconVisibleInMenu(True)
            action.triggered.connect(getattr(self, nom))
        self.clicked.connect(self.mode_occupe)


    def _set_icon(self, nom):
        pix = png(nom)
        self.setIcon(QIcon(pix))
        self.setIconSize(pix.size())

    def mode_exact(self, *args):
        self._set_icon('mode_exact_')
        self.parent.param("calcul_exact", True)
        self.parent.param("ecriture_scientifique", False)

    def mode_approche(self, *args):
        self._set_icon('mode_approche_')
        self.parent.param("calcul_exact", False)
        self.parent.param("ecriture_scientifique", False)

    def mode_scientifique(self, *args):
        self._set_icon('mode_scientifique_')
        self.parent.param("calcul_exact", False)
        self.parent.param("ecriture_scientifique", True)

    def mode_occupe(self):
        self._set_icon('thinking2_')

    def mode_normal(self):
        if self.parent.param("calcul_exact"):
            self.mode_exact()
        elif self.parent.param("ecriture_scientifique"):
            self.mode_scientifique()
        else:
            self.mode_approche()



class PaveNumerique(QWidget):
    def __init__(self, parent):
        QWidget.__init__(self, parent)
        self.parent = parent
        ### Pave numerique de la calculatrice ###
        # On construit le pavé de la calculatrice.
        # Chaque bouton du pavé doit provoquer l'insertion de la commande correspondante.

        self.pave = pave = QVBoxLayout()
#        pave.setSpacing(1)
        boutons = ["2nde", "ans", "ouv", "ferm", "egal", "7", "8", "9", "div", "x", "4", "5", "6", "mul", "y", "1", "2", "3", "minus", "z", "0", "pt", "pow", "plus", "t", "rac", "sin", "cos", "tan", "exp", "i", "pi", "e", "abs", "mod"]
        inserer = ["", "ans()", "(", ")", "=", "7", "8", "9",  "/", "x", "4", "5", "6", "*", "y", "1", "2", "3", "-", "z", "0", ".", "^", "+", "t", "sqrt(", ("sin(", "asin(", "sinus / arcsinus"), ("cos(", "acos(", "cosinus / arccosinus"), ("tan(", "atan(", "tangente / arctangente"), ("exp(", "ln(", "exponentielle / logarithme neperien"), ("i", "cbrt(", "i / racine cubique"), ("pi", "sinh(", "pi / sinus hyperbolique"), ("e", "cosh", "e / cosinus hyperbolique"), ("abs(", "tanh", "valeur absolue / tangente hyperbolique"), (" mod ", "log10(", "modulo / logarithme decimal")]

        self.seconde = False # indique si la touche 2nde est activee.

        self.actions = [self.touche_2nde]

        for i, nom_bouton in enumerate(boutons):
            # On aligne les boutons de la calculatrice par rangées de 5.
            if i%5 == 0:
                self.rangee = rangee = QHBoxLayout()
                rangee.addStretch(1)
                pave.addLayout(rangee)

            # Ensuite, on construit une liste de fonctions, parallèlement à la liste des boutons.
            if i > 0:
                self.actions.append(partial(self.action, commande=inserer[i]))

            bouton = QPushButton()
            pix = png('btn_' + nom_bouton)
            bouton.setIcon(QIcon(pix))
            bouton.setIconSize(pix.size())
            bouton.setFlat(True)
#            bouton.SetBackgroundColour(self.GetBackgroundColour())
            rangee.addWidget(bouton)
            if i%5 == 4:
                rangee.addStretch(2)
            # A chaque bouton, on associe une fonction de la liste.
            bouton.clicked.connect(self.actions[i])
            if type(inserer[i]) == tuple:
                bouton.setToolTip(inserer[i][2])

            self.setLayout(self.pave)

    def touche_2nde(self, event=None):
        self.seconde = not self.seconde
        if self.seconde:
            self.message(u"Touche [2nde] activée.")
        else:
            self.message("")

    def action(self, event=None, commande=''):
        entree = self.parent.entree
        if type(commande) == tuple:
            entree.insert(commande[self.seconde])
        else:
            entree.insert(commande)
        n = entree.cursorPosition()
        entree.setFocus()
        entree.setCursorPosition(n)
        self.seconde = False
        self.parent.message("")



class Options(QWidget):
    def __init__(self, parent):
        QWidget.__init__(self, parent)
        self.parent = parent
        prm = parent.param

        self.ensembles = ('R', 'C')

        ### Liste des options de la calculatrice ###
        self.pave = QVBoxLayout()

        # Chiffres significatifs
        box = QGroupBox(u"Mode calcul approché")
        box_layout = QVBoxLayout()
        box.setLayout(box_layout)

        ligne = QHBoxLayout()
        box_layout.addLayout(ligne)
        ligne.addWidget(QLabel(u"Afficher "))
        self.sc_precision_affichage = sc = QSpinBox(self)
        # param.precision_calcul = 60 par défaut
        sc.setRange(1, 50)
        sc.setValue(prm("precision_affichage"))
        sc.valueChanged.connect(self.EvtPrecisionAffichage)
        ligne.addWidget(sc)
        ligne.addWidget(QLabel(u" chiffre(s) significatif(s)."))
        ligne.addStretch()

        self.pave.addWidget(box)

        # Nombre de décimales
        box = QGroupBox(u"Mode écriture scientifique")
        box_layout = QVBoxLayout()
        box.setLayout(box_layout)

        ligne = QHBoxLayout()
        box_layout.addLayout(ligne)
        ligne.addWidget(QLabel(u"Arrondir les résultats à "))
        self.sc_decimales = sc = QSpinBox(self)
        sc.setRange(0, 11)
        sc.setValue(prm("ecriture_scientifique_decimales"))
        sc.valueChanged.connect(self.EvtDecimales)
        ligne.addWidget(sc)
        ligne.addWidget(QLabel(u" décimale(s)."))
        ligne.addStretch()

        self.pave.addWidget(box)

        box = QGroupBox(u"Copie Automatique")
        box_layout = QVBoxLayout()
        box.setLayout(box_layout)
        # Copie du résultat dans le presse-papier
        ligne = QHBoxLayout()
        box_layout.addLayout(ligne)
        self.cb_copie_automatique = cb = QCheckBox(self)
        cb.setChecked(prm("copie_automatique"))
        cb.stateChanged.connect(self.EvtCopieAutomatique)
        cb.setToolTip(u"Copier automatiquement le résultat du calcul dans le presse-papier.")
        ligne.addWidget(cb)
        ligne.addWidget(QLabel(u"Copie du résultat dans le presse-papier."))
        ligne.addStretch()

        # En mode LaTeX
        ligne = QHBoxLayout()
        box_layout.addLayout(ligne)
        self.cb_copie_automatique_LaTeX = cb = QCheckBox(self)
        cb.setChecked(prm("copie_automatique_LaTeX"))
        cb.setToolTip(u"Copier le résultat du calcul au format LaTeX dans le presse-papier.")
        ligne.addWidget(cb)
        cb.stateChanged.connect(self.EvtCopieAutomatiqueLatex)
        self.st_copie_automatique_LaTeX = st = QLabel(u"Copie au format LaTeX (si possible).")
        ligne.addWidget(st)
        ligne.addStretch()

        self.pave.addWidget(box)
        #~ self.pave.addStretch()

        box = QGroupBox(u"Ensemble de résolution")
        box_layout = QVBoxLayout()
        box.setLayout(box_layout)
        ligne = QHBoxLayout()
        box_layout.addLayout(ligne)
        ligne.addWidget(QLabel(u'Résoudre et factoriser dans '))
        self.cb_ensemble = cb = QComboBox()
        ligne.addWidget(cb)
        cb.addItems((u'R (réels)', u'C (complexes)'))
        cb.setCurrentIndex(self.ensembles.index(prm('ensemble')))
        cb.currentIndexChanged.connect(self.EvtEnsemble)
        self.pave.addWidget(box)
        self.pave.addStretch()

        self.setLayout(self.pave)
        # Pour (dés)activer la ligne "Copie au format LaTeX" au besoin.
        self.EvtCopieAutomatique()

    def EvtPrecisionAffichage(self, event=None):
        val = self.sc_precision_affichage.value()
        self.parent.param("precision_affichage", val)


    def EvtDecimales(self, event=None):
        val = self.sc_decimales.value()
        self.parent.param("ecriture_scientifique_decimales", val)


    def EvtCopieAutomatique(self, event=None):
        valeur = self.cb_copie_automatique.isChecked()
        self.parent.param("copie_automatique", valeur)
        if valeur:
            self.cb_copie_automatique_LaTeX.setEnabled(True)
            self.st_copie_automatique_LaTeX.setEnabled(True)
        else:
            self.cb_copie_automatique_LaTeX.setEnabled(False)
            self.st_copie_automatique_LaTeX.setEnabled(False)


    def EvtCopieAutomatiqueLatex(self, event=None):
        val = self.cb_copie_automatique_LaTeX.isChecked()
        self.parent.param("copie_automatique_LaTeX", val)


    def EvtEnsemble(self, index):
        self.parent.param("ensemble", self.ensembles[index])



class Avance(QWidget):
    def __init__(self, parent):
        QWidget.__init__(self, parent)
        self.parent = parent
        prm = parent.param


        ### Liste des options avancées de la calculatrice ###
        self.pave = QVBoxLayout()

        box = QGroupBox(u"Post-traitement")
        box_layout = QVBoxLayout()
        box.setLayout(box_layout)
        box_layout.addWidget(QLabel(u"Traitement automatiquement du résultat :"))
        self.traitement = QLineEdit()
        traitement = self.parent.param("appliquer_au_resultat") or '_'
        self.traitement.setText(traitement)
        self.traitement.setMinimumWidth(100)
        self.traitement.setToolTip(u"Fonction ou opérations à appliquer automatiquement au résultat (représenté par _). Ex: 'factoriser(_)'.")
        self.traitement.editingFinished.connect(self.EvtAppliquerResultat)
        box_layout.addWidget(self.traitement)
        self.pave.addWidget(box)
        self.setLayout(self.pave)
        self.pave.addStretch()


    def EvtAppliquerResultat(self, event=None):
        val = self.traitement.text().strip()
        if not val:
            self.traitement.setText('_')
            val = None
        elif val == '_':
            val = None
        self.parent.param("appliquer_au_resultat", val)




class OngletsCalc(QTabWidget):
    def __init__(self, parent):
        QTabWidget.__init__(self, parent)
        self.addTab(PaveNumerique(parent), u' Pavé numérique ')
        self.addTab(Options(parent), u'Options')
        self.addTab(Avance(parent), u'Avancé')
        self.setTabPosition(QTabWidget.South)
        self.setStyleSheet("""
        QTabBar::tab:selected {
        background: white;
        border: 1px solid #C4C4C3;
        border-top-color: white; /* same as the pane color */
        border-bottom-left-radius: 4px;
        border-bottom-right-radius: 4px;
        border-top-left-radius: 0px;
        border-top-right-radius: 0px;
        min-width: 8ex;
        padding: 7px;
        }
        QStackedWidget {background:white}
        QTabBar QToolButton {
        background:white;
        border: 1px solid #C4C4C3;
        border-top-color: white; /* same as the pane color */
        border-bottom-left-radius: 4px;
        border-bottom-right-radius: 4px;
        border-top-left-radius: 0px;
        border-top-right-radius: 0px;
        }
        """)





class Calculatrice(Panel_simple):
    titre = u"Calculatrice" # Donner un titre a chaque module

    def __init__(self, *args, **kw):
        Panel_simple.__init__(self, *args, **kw)
        self.interprete = Interprete(calcul_exact = self.param("calcul_exact"),
                                ecriture_scientifique = self.param("ecriture_scientifique"),
                                formatage_OOo = self.param("formatage_OOo"),
                                formatage_LaTeX = self.param("formatage_LaTeX"),
                                ecriture_scientifique_decimales = self.param("ecriture_scientifique_decimales"),
                                precision_calcul = self.param("precision_calcul"),
                                precision_affichage = self.param("precision_affichage"),
                                appliquer_au_resultat = self.param('appliquer_au_resultat'),
                                simpify = True,
                                ensemble=self.param('ensemble'),
                                )

        bouton = BoutonValider(self)
        bouton.setToolTip(u"Laissez appuyé pour changer de mode.")
        self.entree = entree = LigneCommande(self, longueur=550,
                                action=self.affichage_resultat, bouton=bouton)
        entree.setToolTip(u"[Maj]+[Entrée] pour une valeur approchée.")
        self.entree.texte.setContextMenuPolicy(Qt.CustomContextMenu)
        self.entree.texte.customContextMenuRequested.connect(self.EvtMenu)


        self.sizer = sizer = QVBoxLayout()
        sizer.addWidget(entree)
        self.corps = corps = QHBoxLayout()
        sizer.addLayout(corps)
        self.resultats = resultats = QTextEdit(self)
        resultats.setMinimumSize(450, 310)
        resultats.setReadOnly(True)
        corps.addWidget(resultats, 1)
        onglets = OngletsCalc(self)
        corps.addWidget(onglets)
        onglets.setCurrentIndex(self.param('onglet'))
        onglets.currentChanged.connect(self.EvtCurrentChanged)

        self.figure = Figure(figsize=(5,1.3), frameon=True, facecolor="w")
        self.visualisation = FigureCanvas(self.figure)
        self.axes = axes = self.figure.add_axes([0, 0, 1, 1], frameon=False)
        axes.axison = False
        self.pp_texte = axes.text(0.5, 0.5, "", horizontalalignment='center',
                verticalalignment='center', transform = axes.transAxes, size=18)
        self.visualisation.setContextMenuPolicy(Qt.CustomContextMenu)
        self.visualisation.customContextMenuRequested.connect(self.EvtMenuVisualisation)
        sizer.addWidget(self.visualisation)

        self.setLayout(self.sizer)
        self.initialiser()


    def activer(self):
        Panel_simple.activer(self)
        # Actions à effectuer lorsque l'onglet devient actif
        self.entree.setFocus()


    def _sauvegarder(self, fgeo):
        fgeo.contenu["Calculatrice"] = [{}]
        fgeo.contenu["Calculatrice"][0]["Historique"] = [repr(self.entree.historique)]
        fgeo.contenu["Calculatrice"][0]["Affichage"] = [self.resultats.toPlainText()]
        fgeo.contenu["Calculatrice"][0]["Etat_interne"] = [self.interprete.save_state()]



    def _ouvrir(self, fgeo):
        if fgeo.contenu.has_key("Calculatrice"):
            calc = fgeo.contenu["Calculatrice"][0]
            self.initialiser()

            self.entree.historique = eval_safe(calc["Historique"][0])
            resultats = calc["Affichage"][0]
            if resultats:
                resultats += '\n\n'
            self.resultats.setPlainText(resultats)
            self.resultats.moveCursor(QTextCursor.End)
            self.interprete.load_state(calc["Etat_interne"][0])


    def modifier_pp_texte(self, chaine):
        u"""Modifier le résultat affiché en LaTeX (pretty print)."""
        if self.param("latex"):
            # On utilise directement LaTeX pour le rendu
            chaine = "$" + chaine + "$"
        else:
            # On utilise le parser matplotlib.mathtext, moins complet mais bien
            # plus rapide. Certaines adaptations doivent être faites.
            chaine = latex2mathtext(chaine)
        self.pp_texte.set_text(chaine)
        self.visualisation.draw()

    def vers_presse_papier(self, event = None, texte = None):
        if texte is None:
            texte = self.dernier_resultat
        Panel_simple.vers_presse_papier(texte)

    def copier_latex(self, event = None):
        self.vers_presse_papier(texte = self.interprete.latex_dernier_resultat.strip("$"))

    def initialiser(self, event = None):
        self.dernier_resultat = "" # dernier resultat, sous forme de chaine formatee pour l'affichage
        self.entree.initialiser()
        self.interprete.initialiser()
        self.resultats.clear()

    def affichage_resultat(self, commande, **kw):
        # Commandes spéciales:
        if commande in ('clear', 'clear()', 'efface', 'efface()'):
            self.initialiser()
            self.modifier_pp_texte(u"Calculatrice réinitialisée.")
            return

        self.modifie = True
        try:
            try:
                ##self.parent.parent.application.processEvents()
                if kw.get("shift"):
                    self.interprete.calcul_exact = False
                resultat, latex = self.interprete.evaluer(commande)
                if latex == "$?$": # provoque une erreur (matplotlib 0.99.1.1)
                    latex = u"Désolé, je ne sais pas faire..."
            finally:
                self.interprete.calcul_exact = self.param('calcul_exact')
                self.entree.bouton.mode_normal()
            aide = resultat.startswith("\n== Aide sur ")
            if aide:
                latex = ''
            elif not latex:
                latex = resultat
            #LaTeX
            debug("Expression LaTeX: " + latex)
            try:
                try:
                    # Affichage en LaTeX si possible.
                    self.modifier_pp_texte(latex)
                except Exception:
                    print_error()
                    # Sinon, affichage en texte simple.
                    #  `matplotlib.mathtext` est encore loin d'être
                    # pleinement compatible avec LaTeX !
                    self.modifier_pp_texte(resultat)
            except Exception:
                print_error()
                # Si tout a raté... mais ça ne devrait jamais arrivé.
                self.modifier_pp_texte("<Affichage impossible>")
            #Presse-papier
            self.dernier_resultat = resultat
            if self.param("copie_automatique"):
                if self.param("copie_automatique_LaTeX"):
                    self.copier_latex()
                else:
                    self.vers_presse_papier()
            # TextCtrl
            numero = str(len(self.interprete.derniers_resultats))
            # Évite le décalage entre la première ligne et les suivantes (matrices)
            if "\n" in resultat and not aide:
                resultat = "\n" + "\n".join(20*" " + ligne for ligne in resultat.split("\n"))
            self.resultats.moveCursor(QTextCursor.End)
            self.resultats.insertPlainText(u" Calcul n\xb0" + numero + " :   "
                                                        + uu(commande) + u"\n Résultat :"
                                                        + " "*(4+len(numero))
                                                        + resultat + "\n__________________\n\n")
            self.resultats.moveCursor(QTextCursor.End)
            self.message(u"Calcul effectué." + self.interprete.warning)
            self.entree.clear()
#            self.resultats.setCursorPosition(len(self.resultats.plainText()))
#            self.resultats.setFocus()
#            self.resultats.ScrollLines(1)
            self.entree.setFocus()
        except Exception:
            self.message(u"Calcul impossible.")
            self.entree.setFocus()
            if param.debug:
                raise


    def insere(self, event=None, nom='', parentheses=True):
        entree = self.entree
        deb, fin = entree.getSelection()
        if parentheses:
            entree.setCursorPosition(fin)
            entree.insert(")")
            entree.setCursorPosition(deb)
            entree.insert(nom + "(")
            entree.setFocus()
            if deb == fin:
                final = fin + len(nom) + 1
            else:
                final = fin + len(nom) + 2
        else:
            entree.insert(nom)
            final = fin + len(nom)
        entree.setFocus()
        entree.setCursorPosition(final)


    def EvtMenu(self, event):
        menu = QMenu()
        menu.setWindowTitle(u"Fonctions mathématiques")
        debut = True
        for rubrique in __classement__:
            if not debut:
                menu.addSeparator()
            debut = False
            for titre, nom, doc in filter(None, __classement__[rubrique]):
                action = menu.addAction(titre, partial(self.insere, nom=nom, parentheses=(rubrique != "Symboles")))
                # Pas de parenthèses après un symbole.
                action.setToolTip(doc)
        menu.exec_(QCursor.pos())


    def EvtMenuVisualisation(self, event):
        menu = QMenu()
        action = menu.addAction("Copier LaTeX", self.copier_latex)
        action.setToolTip("Copier le code LaTeX dans le presse-papier.")
        menu.exec_(QCursor.pos())

    def EvtCurrentChanged(self, index):
        self.param('onglet', index)

    def param(self, parametre, valeur = no_argument, defaut = False):
        if valeur is not no_argument:
            setattr(self.interprete, parametre, valeur)
        return Panel_simple.param(self, parametre = parametre, valeur = valeur, defaut = defaut)

    def EtatInterne(self, event):
        contenu = self.interprete.save_state()
        h = FenCode(self, u"État interne de l'inteprète", contenu, self.interprete.load_state)
        h.show()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'wxgeometrie/GUI/calibrage.ui'
#
# Created: Thu Jul 10 17:03:16 2014
#      by: PyQt4 UI code generator 4.10.4
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)

class Ui_DialogCalibrage(object):
    def setupUi(self, DialogCalibrage):
        DialogCalibrage.setObjectName(_fromUtf8("DialogCalibrage"))
        DialogCalibrage.resize(510, 271)
        self.verticalLayout = QtGui.QVBoxLayout(DialogCalibrage)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.label_16 = QtGui.QLabel(DialogCalibrage)
        self.label_16.setObjectName(_fromUtf8("label_16"))
        self.verticalLayout.addWidget(self.label_16)
        spacerItem = QtGui.QSpacerItem(20, 40, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.horizontalLayout = QtGui.QHBoxLayout()
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
        self.frame = QtGui.QFrame(DialogCalibrage)
        self.frame.setMinimumSize(QtCore.QSize(300, 30))
        self.frame.setMaximumSize(QtCore.QSize(300, 30))
        self.frame.setStyleSheet(_fromUtf8("background-color: rgb(255, 223, 148);\n"
"border-style: solid;\n"
"border-color: rgb(255, 170, 0);\n"
"border-width: 1px;"))
        self.frame.setFrameShape(QtGui.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtGui.QFrame.Raised)
        self.frame.setObjectName(_fromUtf8("frame"))
        self.horizontalLayout_2 = QtGui.QHBoxLayout(self.frame)
        self.horizontalLayout_2.setObjectName(_fromUtf8("horizontalLayout_2"))
        self.label_2 = QtGui.QLabel(self.frame)
        self.label_2.setStyleSheet(_fromUtf8("color:rgb(255, 170, 0);\n"
"border-style:None;\n"
""))
        self.label_2.setObjectName(_fromUtf8("label_2"))
        self.horizontalLayout_2.addWidget(self.label_2)
        self.horizontalLayout.addWidget(self.frame)
        self.verticalLayout.addLayout(self.horizontalLayout)
        spacerItem1 = QtGui.QSpacerItem(20, 40, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem1)
        self.horizontalLayout_9 = QtGui.QHBoxLayout()
        self.horizontalLayout_9.setObjectName(_fromUtf8("horizontalLayout_9"))
        self.label_17 = QtGui.QLabel(DialogCalibrage)
        self.label_17.setObjectName(_fromUtf8("label_17"))
        self.horizontalLayout_9.addWidget(self.label_17)
        self.longueur = QtGui.QLineEdit(DialogCalibrage)
        self.longueur.setObjectName(_fromUtf8("longueur"))
        self.horizontalLayout_9.addWidget(self.longueur)
        self.label_18 = QtGui.QLabel(DialogCalibrage)
        self.label_18.setObjectName(_fromUtf8("label_18"))
        self.horizontalLayout_9.addWidget(self.label_18)
        spacerItem2 = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout_9.addItem(spacerItem2)
        self.verticalLayout.addLayout(self.horizontalLayout_9)
        spacerItem3 = QtGui.QSpacerItem(20, 40, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem3)
        self.line = QtGui.QFrame(DialogCalibrage)
        self.line.setFrameShape(QtGui.QFrame.HLine)
        self.line.setFrameShadow(QtGui.QFrame.Sunken)
        self.line.setObjectName(_fromUtf8("line"))
        self.verticalLayout.addWidget(self.line)
        self.buttonBox = QtGui.QDialogButtonBox(DialogCalibrage)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Cancel|QtGui.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName(_fromUtf8("buttonBox"))
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(DialogCalibrage)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("accepted()")), DialogCalibrage.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("rejected()")), DialogCalibrage.close)
        QtCore.QMetaObject.connectSlotsByName(DialogCalibrage)

    def retranslateUi(self, DialogCalibrage):
        DialogCalibrage.setWindowTitle(_translate("DialogCalibrage", "Calibrage de l\'écran", None))
        self.label_16.setText(_translate("DialogCalibrage", "<html><head/><body><p>Pour calibrer votre écran mesurez avec une règle la barre orange suivante.</p><p>Reportez ensuite ci-dessous sa <span style=\" font-weight:600;\">longueur</span> en cm, pour que la résolution<br/>soit automatiquement calculée.</p></body></html>", None))
        self.label_2.setText(_translate("DialogCalibrage", "Mesurez moi !", None))
        self.label_17.setText(_translate("DialogCalibrage", "Longueur mesurée : ", None))
        self.label_18.setText(_translate("DialogCalibrage", "cm", None))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#   Mathlib 2 (sympy powered) #
##--------------------------------------#######
#WxGeometrie
#Dynamic geometry, graph plotter, and more for french mathematic teachers.
#Copyright (C) 2005-2013  Nicolas Pourcelot
#
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode

## Les fonctions qui doivent être accessibles par l'utilisateur final
## "from end_user_functions import *" doit donner un résultat propre.
from ..pylib import OrderedDict

__classement__ = OrderedDict((
    (u"Algèbre", []),
    (u"Analyse", []),
    (u"Symboles", []),
    (u"Arithmétique", []),
    (u"Statistiques", []),
    (u"Matrices", []),
    (u"Divers", []),
##    (u"LaTeX", []),
))

del OrderedDict

from .universal_functions import abs, acos, asin, atan, ceil, cos, cosh, exp, floor,\
                                 ln, log, sin, sinh, sqrt, tan, tanh, arg


from sympy import LambertW, gamma, Sum, Integral

# Alias
arcsin = asin
arccos = acos
arctan = atan

from .custom_functions import derivee, moyenne, variance, ecart_type, \
                               covariance, linreg, aide, normal, binomial, \
                               inv_normal, fluctu, confiance


__classement__[u"Statistiques"].append((u"Moyenne", "moyenne", u"Moyenne de la série, éventuellement coefficientée."
                                     u" Ex: moyenne((12; 13; 15); (1; 1; 2))"))
__classement__[u"Statistiques"].append((u"Variance", "variance", u"Variance de la série, éventuellement coefficientée."
                                     u" Ex: variance((12; 13; 15); (1; 1; 2))"))
__classement__[u"Statistiques"].append((u"Écart-type", "ecart_type",
                        u"Écart-type de la série, éventuellement coefficientée."
                                     u" Ex: ecart_type((12; 13; 15); (1; 1; 2))"))
__classement__[u"Statistiques"].append((u"Covariance", "covariance", u"Covariance de deux séries."
                                     u" Ex: covariance((1;2;3);(2;5;7))"))
__classement__[u"Statistiques"].append((u"Droite de régression", "linreg",
                u"Retourne (a, b) pour une droite d'ajustement d'équation y=ax+b"
                                     u" (moindres carrés). Ex: linreg((1;2;3);(2;5;7))"))
__classement__[u"Statistiques"].append((u"Loi normale", "normal",
                u"Usage: normal(a, b, mu, sigma). Retourne P(a \u2264 X \u2264 b), où "
                u"X suit une loi normale d'espérance mu et de variance sigma²."))
__classement__[u"Statistiques"].append((u"Loi binomiale", "binomial",
                u"Usage: normal(a, b, n, p). Retourne P(a \u2264 X \u2264 b), où "
                u"X suit une loi binomiale de paramètres n et p."))
__classement__[u"Statistiques"].append((u"Inverser la loi normale", "inv_normal",
                u"Usage: inv_normal(p). Retourne a tel que P(X \u2264 a)=p, où "
                u"X suit une loi normale centrée réduite."))
__classement__[u"Statistiques"].append((u"Intervalle de fluctuation", "fluctu",
                u"Usage: fluctu(p, n, seuil). Retourne l'intervalle de fluctuation "
                u"pour un échantillon de taille n, et une proportion p."))
__classement__[u"Statistiques"].append((u"Intervalle de confiance", "confiance",
                u"Usage: estim95(f, n). Retourne l'intervalle de confiance "
                u"pour un échantillon de taille n, et une fréquence f."))

from .sympy_functions import expand
developper = developpe = expand
__classement__[u"Algèbre"].append((u"Développe", "developpe", u"Développer une expression."))

from .sympy_functions import factor
factoriser = factorise = factor
__classement__[u"Algèbre"].append((u"Factorise", "factorise", u"Factoriser une expression."))

from .custom_functions import canonique
__classement__[u"Algèbre"].append((u"Forme canonique", "canonique", u"Mettre sous forme canonique un polynôme de degré 2."))

from .custom_functions import discriminant
__classement__[u"Algèbre"].append((u"Discriminant", "discriminant", u"Calcule le discriminant d'un polynôme de degré 2."))

from .sympy_functions import together
rassembler = rassemble = together
__classement__[u"Algèbre"].append((u"Une seule fraction", "rassemble", u"Mettre sur le même dénominateur."))


from .sympy_functions import evalf
evaluer = evalue = evalf
__classement__[u"Algèbre"].append((u"Évalue", "evalue", u"Calculer une valeur approchée."))

from .solvers import resoudre
solve = resous = resoudre
__classement__[u"Algèbre"].append((u"Résous", "resous", u"Résoudre une (in)équation ou un système. ex: resous(2x-3>0 et 3x+1<=0)"))

__classement__[u"Algèbre"].append(None)

#~ from .sympy_functions import cfactor
#~ cfactoriser = cfactorise = cfactor
#~ __classement__[u"Algèbre"].append((u"Factorise dans C", "cfactorise", u"Factoriser une expression dans le corps des complexes."))

from sympy import conjugate
conjug = conjugue = conjuguer = conjugate
__classement__[u"Algèbre"].append((u"Conjuguer", "conjug", u"Calculer le conjugué d'un nombre complexe."))

__classement__[u"Algèbre"].append(None)

from .sympy_functions import sum
somme = sum
__classement__[u"Algèbre"].append((u"Somme", "somme", u"Calculer une somme."))

from .sympy_functions import product
produit = product
__classement__[u"Algèbre"].append((u"Produit", "produit", u"Calculer un produit."))

__classement__[u"Algèbre"].append(None)

from .sympy_functions import mat, transpose, det, vecteurs_propres, valeurs_propres
Matrice = Matrix = Mat = matrix = matrice = mat
vep = vecteurs_propres
vap = valeurs_propres
determinant = det
tr = transposee = transpose
__classement__[u"Matrices"].append((u"Matrice", "mat", u"Créer une matrice. ex: mat([[1;2];[3;4]]); mat(4;4;0); mat(4;4;2*li+3*co)"))
__classement__[u"Matrices"].append((u"Transpose", "tr", u"Transpose une matrice. ex: tr(M)"))
__classement__[u"Matrices"].append((u"Déterminant", "det", u"Calcul le déterminant d'une matrice. ex: det(M)"))
__classement__[u"Matrices"].append((u"Valeurs propres", "vap", u"Valeurs propres d'une matrice avec ordre de multiplicité. ex: vap(M)"))
__classement__[u"Matrices"].append((u"Vecteurs propres", "vep", u"Base de vecteurs propres d'une matrice. ex: vep(M)"))


from sympy import isprime
premier = isprime
__classement__[u"Arithmétique"].append((u"Premier ?", "premier", u"Tester si un nombre est premier."))

from .sympy_functions import divisors
diviseurs = divisors
__classement__[u"Arithmétique"].append((u"Diviseurs", "diviseurs", u"Chercher les diviseurs d'un nombre."))

# Alias
facteurs = factors = factor
__classement__[u"Arithmétique"].append((u"Facteurs premiers", "facteurs", u"Décomposer un nombre en facteurs premiers."))

from .custom_functions import pgcd
gcd = pgcd
__classement__[u"Arithmétique"].append((u"PGCD", "pgcd", u"Calculer le PGCD de plusieurs entiers."))

from .custom_functions import ppcm
lcm = ppcm
__classement__[u"Arithmétique"].append((u"PPCM", "ppcm", u"Calculer le PPCM de plusieurs entiers."))

__classement__[u"Arithmétique"].append(None)

from .custom_functions import jhms
hms = jhms
__classement__[u"Divers"].append((u"Conversion j-h-min-s", "jhms", u"Convertir un temps en secondes en jours, heures, minutes et secondes."))

from .custom_functions import deg
__classement__[u"Divers"].append((u"Conversion en degrés", "deg", u"Convertir un angle de radians en degrés. (Ex: deg(pi/3), ou pi/3>>deg)."))

from .custom_functions import frac
__classement__[u"Divers"].append((u"Convertir en fraction", "frac", u"Convertir un nombre décimal en fraction."))

ent = floor
__classement__[u"Divers"].append((u"Partie entière", "ent", u"Partie entière d'un nombre (ex: -2,4 -> -3)."))

from .custom_functions import arrondir
round = arrondi = arrondis = arrondir
__classement__[u"Divers"].append((u"Arrondi", "arrondi", u"Arrondit à n chiffres après la virgule-3)."))


from sympy import factorial
factoriel = factorial
__classement__[u"Arithmétique"].append((u"Factoriel", "factoriel", u"factoriel(n) vaut 1*2*3*4*5*...*n."))

from sympy import ff
arrangements = nAr = nA = ff
__classement__[u"Arithmétique"].append((u"Arrangements", "nAr", u"Nombre d'arrangements."))

from sympy import binomial as binome
combinaisons = combi = nCr = nC = binome
__classement__[u"Arithmétique"].append((u"Combinaisons", "nCr", u"Nombre de combinaisons."))

from .sympy_functions import limit
lim = limite = limit
__classement__[u"Analyse"].append((u"Limite", "limite", u"Calculer une limite. Ex: 'limite(1/x; x; 0+)', 'limite(x^2; x; -oo)'"))

from .sympy_functions import diff
derive = diff
__classement__[u"Analyse"].append((u"Dérive", "derive", u"Dériver une expression. Ex: 'derive(2x+7; x)' et 'derive(2x+7; x; 2)' (dérivée seconde)"))

from .sympy_functions import integrate
integrale = integral = integre = integrate
__classement__[u"Analyse"].append((u"Intègre", "integre", u"Intégrer une expression. Ex: 'integre(x+1; x)' et 'integre(x+1; (x; -1; 1))"))

from .sympy_functions import series
taylor = series
__classement__[u"Analyse"].append((u"Taylor", "taylor", u"Développement limité. Ex: 'taylor(cos(x); x; 0; 5)'"))

from sympy import pi
__classement__[u"Symboles"].append((u"pi", "pi", u"Le nombre pi (3,14...)."))

from sympy import oo
__classement__[u"Symboles"].append((u"oo", "oo", u"L'infini positif."))

from sympy import E
e = E
__classement__[u"Symboles"].append((u"e", "e", u"Le nombre e = exp(1)."))

from sympy import I
i = I
__classement__[u"Symboles"].append((u"i", "i", u"Le nombre imaginaire i."))

##from custom_functions import pstfunc
##__classement__[u"LaTeX"].append((u"Fonction pst-plot", "pstfunc", u"Conversion d'une formule pst-plot en fonction."))


__classement__[u"Divers"].append(None)

from .custom_functions import pstfunc
__classement__[u"Divers"].append((u"Import d'une fonction \\psplot", "pstfunc", u'Pour utilisateurs de LaTeX. Ex: pstfunc("4 1 2.71818 0.039 x mul neg exp sub div")'))

__classement__[u"Divers"].append(None)

__classement__[u"Divers"].append((u"Aide", "aide", u"Fournit un complément d'aide sur certaines fonctions. Ex: aide(linreg)"))
help = aide


from .graphes import Graph
Graphe = Graph

#~ class Exp1_bis(sympy.core.numbers.Exp1):
    #~ def tostr(self, level=0):
        #~ return 'e'

#~ e = Exp1_bis()

#~ del Exp1_bis

#~ class ImaginaryUnit_bis(sympy.core.numbers.ImaginaryUnit):
    #~ def tostr(self, level=0):
        #~ return 'i'

#~ i = ImaginaryUnit_bis()

#~ del ImaginaryUnit_bis



# -*- coding: utf-8 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

NOMPROG = u"Géophar"
NOMPROG2 = 'Geophar' # sans accent

# Numéro de version (et date de sa sortie)
version = '16.08.4'
date_version = (2016, 11, 6)
git = 'v16.08.3-2-gd654f7a'


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'wxgeometrie/GUI/reperage.ui'
#
# Created: Thu Jul 10 18:02:21 2014
#      by: PyQt4 UI code generator 4.10.4
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)

class Ui_DialogReperage(object):
    def setupUi(self, DialogReperage):
        DialogReperage.setObjectName(_fromUtf8("DialogReperage"))
        DialogReperage.resize(444, 493)
        self.verticalLayout = QtGui.QVBoxLayout(DialogReperage)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.groupBoxRepere = QtGui.QGroupBox(DialogReperage)
        self.groupBoxRepere.setObjectName(_fromUtf8("groupBoxRepere"))
        self.verticalLayout_3 = QtGui.QVBoxLayout(self.groupBoxRepere)
        self.verticalLayout_3.setObjectName(_fromUtf8("verticalLayout_3"))
        self.horizontalLayout = QtGui.QHBoxLayout()
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
        self.label = QtGui.QLabel(self.groupBoxRepere)
        self.label.setObjectName(_fromUtf8("label"))
        self.horizontalLayout.addWidget(self.label)
        self.repere_O = QtGui.QLineEdit(self.groupBoxRepere)
        self.repere_O.setObjectName(_fromUtf8("repere_O"))
        self.horizontalLayout.addWidget(self.repere_O)
        self.label_2 = QtGui.QLabel(self.groupBoxRepere)
        self.label_2.setObjectName(_fromUtf8("label_2"))
        self.horizontalLayout.addWidget(self.label_2)
        self.repere_I = QtGui.QLineEdit(self.groupBoxRepere)
        self.repere_I.setObjectName(_fromUtf8("repere_I"))
        self.horizontalLayout.addWidget(self.repere_I)
        self.label_3 = QtGui.QLabel(self.groupBoxRepere)
        self.label_3.setObjectName(_fromUtf8("label_3"))
        self.horizontalLayout.addWidget(self.label_3)
        self.repere_J = QtGui.QLineEdit(self.groupBoxRepere)
        self.repere_J.setObjectName(_fromUtf8("repere_J"))
        self.horizontalLayout.addWidget(self.repere_J)
        self.label_4 = QtGui.QLabel(self.groupBoxRepere)
        self.label_4.setObjectName(_fromUtf8("label_4"))
        self.horizontalLayout.addWidget(self.label_4)
        spacerItem = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.verticalLayout.addWidget(self.groupBoxRepere)
        self.groupBoxGraduations = QtGui.QGroupBox(DialogReperage)
        self.groupBoxGraduations.setObjectName(_fromUtf8("groupBoxGraduations"))
        self.verticalLayout_4 = QtGui.QVBoxLayout(self.groupBoxGraduations)
        self.verticalLayout_4.setObjectName(_fromUtf8("verticalLayout_4"))
        self.horizontalLayout_2 = QtGui.QHBoxLayout()
        self.horizontalLayout_2.setObjectName(_fromUtf8("horizontalLayout_2"))
        self.label_5 = QtGui.QLabel(self.groupBoxGraduations)
        self.label_5.setObjectName(_fromUtf8("label_5"))
        self.horizontalLayout_2.addWidget(self.label_5)
        self.gradu_x = QtGui.QLineEdit(self.groupBoxGraduations)
        self.gradu_x.setObjectName(_fromUtf8("gradu_x"))
        self.horizontalLayout_2.addWidget(self.gradu_x)
        self.label_7 = QtGui.QLabel(self.groupBoxGraduations)
        self.label_7.setObjectName(_fromUtf8("label_7"))
        self.horizontalLayout_2.addWidget(self.label_7)
        spacerItem1 = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.verticalLayout_4.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3 = QtGui.QHBoxLayout()
        self.horizontalLayout_3.setObjectName(_fromUtf8("horizontalLayout_3"))
        self.label_6 = QtGui.QLabel(self.groupBoxGraduations)
        self.label_6.setObjectName(_fromUtf8("label_6"))
        self.horizontalLayout_3.addWidget(self.label_6)
        self.gradu_y = QtGui.QLineEdit(self.groupBoxGraduations)
        self.gradu_y.setObjectName(_fromUtf8("gradu_y"))
        self.horizontalLayout_3.addWidget(self.gradu_y)
        self.label_8 = QtGui.QLabel(self.groupBoxGraduations)
        self.label_8.setObjectName(_fromUtf8("label_8"))
        self.horizontalLayout_3.addWidget(self.label_8)
        spacerItem2 = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem2)
        self.verticalLayout_4.addLayout(self.horizontalLayout_3)
        self.verticalLayout.addWidget(self.groupBoxGraduations)
        self.groupBoxEchelle = QtGui.QGroupBox(DialogReperage)
        self.groupBoxEchelle.setObjectName(_fromUtf8("groupBoxEchelle"))
        self.verticalLayout_5 = QtGui.QVBoxLayout(self.groupBoxEchelle)
        self.verticalLayout_5.setObjectName(_fromUtf8("verticalLayout_5"))
        self.horizontalLayout_4 = QtGui.QHBoxLayout()
        self.horizontalLayout_4.setObjectName(_fromUtf8("horizontalLayout_4"))
        self.label_9 = QtGui.QLabel(self.groupBoxEchelle)
        self.label_9.setObjectName(_fromUtf8("label_9"))
        self.horizontalLayout_4.addWidget(self.label_9)
        self.echelle_x = QtGui.QLineEdit(self.groupBoxEchelle)
        self.echelle_x.setObjectName(_fromUtf8("echelle_x"))
        self.horizontalLayout_4.addWidget(self.echelle_x)
        self.label_12 = QtGui.QLabel(self.groupBoxEchelle)
        self.label_12.setObjectName(_fromUtf8("label_12"))
        self.horizontalLayout_4.addWidget(self.label_12)
        spacerItem3 = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem3)
        self.verticalLayout_5.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_5 = QtGui.QHBoxLayout()
        self.horizontalLayout_5.setObjectName(_fromUtf8("horizontalLayout_5"))
        self.label_13 = QtGui.QLabel(self.groupBoxEchelle)
        self.label_13.setObjectName(_fromUtf8("label_13"))
        self.horizontalLayout_5.addWidget(self.label_13)
        self.echelle_y = QtGui.QLineEdit(self.groupBoxEchelle)
        self.echelle_y.setObjectName(_fromUtf8("echelle_y"))
        self.horizontalLayout_5.addWidget(self.echelle_y)
        self.label_10 = QtGui.QLabel(self.groupBoxEchelle)
        self.label_10.setObjectName(_fromUtf8("label_10"))
        self.horizontalLayout_5.addWidget(self.label_10)
        spacerItem4 = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem4)
        self.verticalLayout_5.addLayout(self.horizontalLayout_5)
        self.rapport = QtGui.QCheckBox(self.groupBoxEchelle)
        self.rapport.setObjectName(_fromUtf8("rapport"))
        self.verticalLayout_5.addWidget(self.rapport)
        self.verticalLayout.addWidget(self.groupBoxEchelle)
        self.groupBoxResolution = QtGui.QGroupBox(DialogReperage)
        self.groupBoxResolution.setObjectName(_fromUtf8("groupBoxResolution"))
        self.verticalLayout_2 = QtGui.QVBoxLayout(self.groupBoxResolution)
        self.verticalLayout_2.setObjectName(_fromUtf8("verticalLayout_2"))
        self.label_15 = QtGui.QLabel(self.groupBoxResolution)
        self.label_15.setObjectName(_fromUtf8("label_15"))
        self.verticalLayout_2.addWidget(self.label_15)
        self.horizontalLayout_7 = QtGui.QHBoxLayout()
        self.horizontalLayout_7.setObjectName(_fromUtf8("horizontalLayout_7"))
        self.label_14 = QtGui.QLabel(self.groupBoxResolution)
        self.label_14.setObjectName(_fromUtf8("label_14"))
        self.horizontalLayout_7.addWidget(self.label_14)
        self.resolution = QtGui.QLineEdit(self.groupBoxResolution)
        self.resolution.setObjectName(_fromUtf8("resolution"))
        self.horizontalLayout_7.addWidget(self.resolution)
        self.label_11 = QtGui.QLabel(self.groupBoxResolution)
        self.label_11.setObjectName(_fromUtf8("label_11"))
        self.horizontalLayout_7.addWidget(self.label_11)
        spacerItem5 = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem5)
        self.calibrage = QtGui.QPushButton(self.groupBoxResolution)
        self.calibrage.setObjectName(_fromUtf8("calibrage"))
        self.horizontalLayout_7.addWidget(self.calibrage)
        self.verticalLayout_2.addLayout(self.horizontalLayout_7)
        self.verticalLayout.addWidget(self.groupBoxResolution)
        self.line = QtGui.QFrame(DialogReperage)
        self.line.setFrameShape(QtGui.QFrame.HLine)
        self.line.setFrameShadow(QtGui.QFrame.Sunken)
        self.line.setObjectName(_fromUtf8("line"))
        self.verticalLayout.addWidget(self.line)
        self.boutons = QtGui.QDialogButtonBox(DialogReperage)
        self.boutons.setOrientation(QtCore.Qt.Horizontal)
        self.boutons.setStandardButtons(QtGui.QDialogButtonBox.Apply|QtGui.QDialogButtonBox.Cancel|QtGui.QDialogButtonBox.Ok|QtGui.QDialogButtonBox.Reset)
        self.boutons.setObjectName(_fromUtf8("boutons"))
        self.verticalLayout.addWidget(self.boutons)

        self.retranslateUi(DialogReperage)
        QtCore.QObject.connect(self.boutons, QtCore.SIGNAL(_fromUtf8("accepted()")), DialogReperage.accept)
        QtCore.QObject.connect(self.boutons, QtCore.SIGNAL(_fromUtf8("rejected()")), DialogReperage.close)
        QtCore.QMetaObject.connectSlotsByName(DialogReperage)

    def retranslateUi(self, DialogReperage):
        DialogReperage.setWindowTitle(_translate("DialogReperage", "Personnaliser le repère", None))
        self.groupBoxRepere.setTitle(_translate("DialogReperage", "Repère", None))
        self.label.setText(_translate("DialogReperage", "Entrer le repère : (", None))
        self.repere_O.setToolTip(_translate("DialogReperage", "<html><head/><body><p>Légende à afficher pour l\'origine du repère.</p></body></html>", None))
        self.label_2.setText(_translate("DialogReperage", ";", None))
        self.repere_I.setToolTip(_translate("DialogReperage", "<html><head/><body><p>Légende à afficher pour le 2e point du repère (abscisse).</p><p><span style=\" font-weight:600;\">Nota : </span>si une lettre en minuscule est rentrée, elle sera comprise comme un vecteur.</p></body></html>", None))
        self.label_3.setText(_translate("DialogReperage", ";", None))
        self.repere_J.setToolTip(_translate("DialogReperage", "<html><head/><body><p>Légende à afficher pour le 3e point du repère (ordonnée). </p><p><span style=\" font-weight:600;\">Nota : </span>si une lettre en minuscule est rentrée, elle sera comprise comme un vecteur.</p></body></html>", None))
        self.label_4.setText(_translate("DialogReperage", ")", None))
        self.groupBoxGraduations.setTitle(_translate("DialogReperage", "Graduations", None))
        self.label_5.setText(_translate("DialogReperage", "Graduer en abscisse toutes les ", None))
        self.label_7.setText(_translate("DialogReperage", "unités", None))
        self.label_6.setText(_translate("DialogReperage", "Graduer en ordonnée toutes les ", None))
        self.label_8.setText(_translate("DialogReperage", "unités", None))
        self.groupBoxEchelle.setTitle(_translate("DialogReperage", "Échelle", None))
        self.label_9.setText(_translate("DialogReperage", "Unité en abscisse :", None))
        self.label_12.setText(_translate("DialogReperage", "cm", None))
        self.label_13.setText(_translate("DialogReperage", "Unité en ordonnée :", None))
        self.label_10.setText(_translate("DialogReperage", "cm", None))
        self.rapport.setToolTip(_translate("DialogReperage", "<html><head/><body><p>Cochez cette case pour conserver un rapport abscisse/ordonnée constant lorsque vous zoomez ou redimensionnez la fenêtre.</p></body></html>", None))
        self.rapport.setText(_translate("DialogReperage", "Conserver ce rapport abscisse/ordonnée lors d\'un zoom", None))
        self.groupBoxResolution.setTitle(_translate("DialogReperage", "Résolution", None))
        self.label_15.setText(_translate("DialogReperage", "<html><head/><body><p>Ajuster ce paramètre en fonction de votre écran.<br/><span style=\" font-style:italic;\">Ceci permet d\'affiner l\'affichage pour qu\'un centimètre <br/>mesure vraiment un centimètre sur votre écran !</span></p></body></html>", None))
        self.label_14.setText(_translate("DialogReperage", "Résolution", None))
        self.resolution.setToolTip(_translate("DialogReperage", "<html><head/><body><p>Entrez ici la résolution de votre moniteur en dpi. Exemple : 96.</p></body></html>", None))
        self.label_11.setText(_translate("DialogReperage", "dpi", None))
        self.calibrage.setToolTip(_translate("DialogReperage", "<html><head/><body><p>Afficher un objet à mesurer pour calibrer l\'écran.</p></body></html>", None))
        self.calibrage.setText(_translate("DialogReperage", "Outil de calibrage", None))




# -*- coding: utf-8 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

afficher_axes = False
afficher_quadrillage = False
placement_probabilites = ['dessus', 'longe', 'decale'][1]



# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

from random import random, randint as _randint
from os.path import split, realpath, abspath
import sys

from wxgeometrie.param import tolerance as EPSILON

_module_path = split(realpath(sys._getframe().f_code.co_filename))[0]
ROOTDIR = abspath(_module_path + '/..') # /.../nom_du_projet/
WXGEODIR = ROOTDIR + '/wxgeometrie'

def randint(a, b = None):
    if b is None:
        b = a
        a = 0
    return _randint(a, b)

def rand():
    return randint(50) - randint(50) + random()

def assertAlmostEqual(x, y):
    if type(x) == type(y) and isinstance(x, (tuple, list)):
        for x_elt, y_elt in zip(x, y):
            assertAlmostEqual(x_elt, y_elt)
    else:
        TEST = abs(y - x) < EPSILON
        if not TEST:
            print('''
--------------
 *** FAIL ***
-> Output:
%s
-> Expected:
%s
--------------
''' %(repr(x), repr(y)))
            assert TEST

def assertNotAlmostEqual(x, y):
    # TODO: define test for tuple
    TEST = abs(y - x) > EPSILON
    if not TEST:
        print x,  "==",  y
    assert TEST

def assertEqual(x, y):
    def yellow(s):
        return '\033[0;33m' + s + '\033[0m'
    if x != y:
        rx = repr(x)
        ry = repr(y)
        if isinstance(x, basestring) and isinstance(y, basestring):
            for i in range(min(len(rx), len(ry))):
                if rx[i] != ry[i]:
                    break
            ry = ry[:i] + yellow(ry[i:])

        print('''
--------------
 *** FAIL ***
-> Output:
%s
-> Expected:
%s
--------------
''' %(rx, ry))
    assert (x == y)

assertEq = assertEqual

def assertRaises(error, f, *args, **kw):
    try:
        f(*args, **kw)
    except Exception as e:
        assert type(e) == error
    else:
        raise AssertionError, '%s should be raised.' %type(e).__name__



# -*- coding: utf-8 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)
from __future__ import with_statement

#    .-----------------------------------------------------.
#    |    Exercices : inéquations produits et quotients    |
#    '-----------------------------------------------------'
#    Géophar
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2013  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import re
from random import randint
from itertools import chain
from functools import partial

from PyQt4.QtGui import (QVBoxLayout, QLabel, QPushButton, QHBoxLayout,
                         QMessageBox, QTextEdit, QColor)

from sympy import S, solve, gcd
from sympy.core.sympify import SympifyError

from ...GUI.menu import MenuBar
from ...GUI.panel import Panel_API_graphique
from ...GUI.proprietes_objets import Proprietes
from ...GUI.qtlib import BusyCursor
from ...geolib import Segment, Texte, Point, Champ
from ...geolib.routines import nice_str
from ...pylib import OrderedDict, print_error
from ...mathlib.parsers import convertir_en_latex, NBR, NBR_SIGNE
from ... import param


# TODO:
# - ajouter un bonus substantiel si les résultats sont donnés
#   sous forme de fraction simplifiée.
# - possibilité de définir facilement les niveaux (fichier texte externe ?)
# - gestion des carrés, des nombres seuls



class TabMenuBar(MenuBar):
    def __init__(self, panel):
        MenuBar.__init__(self, panel)

        self.ajouter(u"Fichier", [u"Recommencer", u"Recommencer au niveau 0.", u"Ctrl+N", panel.reinitialiser],
                    [u"ouvrir"],
                    [u"enregistrer"], [u"enregistrer_sous"], [u"exporter"],
                    [u"exporter&sauver"], None, [u"imprimer"], [u"presse-papier"],
                    None, [u"proprietes"], None, ["fermer"], ["quitter"])
        self.ajouter(u"Editer", ["annuler"], ["refaire"], ["modifier"], ["supprimer"])
        self.ajouter(u"Affichage", ["onglet"], ["plein_ecran"], None, ["zoom_texte"], ["zoom_ligne"], ["zoom_general"])
        self.ajouter(u"Outils", [u"options"])
        self.ajouter(u"avance1")
        self.ajouter(u"?")



class ExercicesTableauxSignes(Panel_API_graphique):

    titre = u"Tableaux de signes" # Donner un titre a chaque module

    def __init__(self, *args, **kw):
        Panel_API_graphique.__init__(self, *args, **kw)
        ##"(nx+n)(nx+n)|nx+n"
        # numerateur|denominateur
        # facteur1, facteur2,...|facteur1, facteur2,...
        # n: entier, q: fraction, d: decimal
        self.canvas.fixe = True

        self.entrees = QVBoxLayout()
        self.entrees.addSpacing(30)

        self.panneau = QLabel('')
        self.panneau.setStyleSheet(
            """QLabel { padding: 10px; border-width: 2px; border-style:solid;
            border-radius: 5px; border-color:%s; background-color: %s }"""
            %(QColor(30, 144, 255).name(), QColor(176, 226, 255).name())
                        )
        self.entrees.addWidget(self.panneau)

        self.entrees.addStretch()
        self.felicitations = QLabel('')
        self.entrees.addWidget(self.felicitations)

        self.entrees.addSpacing(30)
        self.btn_niveau = QPushButton(u"Niveau suivant", self)
        self.btn_niveau.clicked.connect(self.niveau_suivant)
        self.entrees.addWidget(self.btn_niveau)
        self.entrees.addSpacing(50)

        self.sizer = QHBoxLayout()
        self.sizer.addWidget(self.canvas, 1)
        self.sizer.addLayout(self.entrees, 0.2)
        self.finaliser(contenu=self.sizer)

        # Ne pas éditer les champs/textes avec [Entrée]
        self.canvas.editeur.active = False

        self.niveaux = ["n*x+z", "-n*x+q", "1|q*x+z", "z*x+z,z*x+z", "z*x+q|z*x+z",
                        "z*x+z,z*x+z|z*x+z", "z*x+z,z*x+z|z*x+z,z*x+z",
                        "-n,z*x+z|-x,(z*x+z)**2"]

        self.reinitialiser()


    def reinitialiser(self):
        # Ne pas éditer les objets par un clic droit
        self.canvas.edition_par_clic_droit = False
        self.score = 0
        self.niveau = -1
        self.erreurs = 0
        self.canvas.afficher_axes = False
        self.canvas.afficher_quadrillage = False
        self.niveau_suivant()


    def niveau_suivant(self, niveau=None):
        if niveau in (None, False):
            # None ou False (False est renvoyé par Qt via QAbstractBouton.clicked)
            self.niveau += 1
        else:
            self.niveau = niveau
        self.btn_niveau.setEnabled(False)
        self.felicitations.setStyleSheet(
            """QLabel {background-color: white; padding: 5px; border-radius: 5px;
            color:white;}""")
        self.fermer_feuilles()
        self.update_panneau()
        self.pattern = self.niveaux[self.niveau]
        self.generer_expression()

    n = niveau_suivant

    def update_panneau(self):
        self.panneau.setText((u"<p><b><i>Niveau :</i> %s</b></p>" % self.niveau) +
                                 (u"<p><b><i>Points :</i> %s</b></p>" % self.score) +
                                 (u"<p><i>Erreurs :</i> %s</p>" % self.erreurs))
        champs = self.feuille_actuelle.objets.lister(type=Champ)
        if champs and all(obj.correct for obj in champs):
            if self.niveau + 1 < len(self.niveaux):
                self.btn_niveau.setEnabled(True)
                self.btn_niveau.setFocus(True)
                self.felicitations.setText(u'<p><b>Félicitations !</b></p>' +
                                           u'<p>Passer au niveau %s</p>' %(self.niveau + 1))
                self.felicitations.setStyleSheet(
                    """QLabel {background-color: %s; padding: 5px;
                       border-radius: 5px;
                       color:white;}""" %QColor(255, 153, 0).name())

            else:
                self.felicitations.setText(u'<p><b>Félicitations !</b></p>' +
                                           u'<p>Dernier niveau terminé !</p>')
                self.felicitations.setStyleSheet(
                    """QLabel {background-color: %s; padding: 5px; border-radius: 5px;
                    color:white;}""" %QColor(102, 205, 0).name())


    def _sauvegarder(self, fgeo, feuille = None):
        Panel_API_graphique._sauvegarder(self, fgeo, feuille)
        fgeo.contenu[u"niveau"] = [str(self.niveau)]
        fgeo.contenu[u"expression"] = [self.raw_expression]
        fgeo.contenu[u"score"] = [str(self.score)]
        fgeo.contenu[u"erreurs"] = [str(self.erreurs)]

    def _ouvrir(self, fgeo):
        # Il ne doit y avoir qu'une seule feuille ouverte à la fois.
        # XXX: intégrer cette fonctionnalité directement au Panel.
        self.fermer_feuilles()
        Panel_API_graphique._ouvrir(self, fgeo)
        if fgeo.contenu.has_key(u"expression"):
            self.generer_expression(expr=fgeo.contenu[u"expression"][0])
            ##self.dessiner_tableau()
        if fgeo.contenu.has_key(u"niveau"):
            self.niveau = int(fgeo.contenu[u"niveau"][0])
        if fgeo.contenu.has_key(u"score"):
            self.score = int(fgeo.contenu[u"score"][0])
        if fgeo.contenu.has_key(u"erreurs"):
            self.erreurs = int(fgeo.contenu[u"erreurs"][0])
        self.update_panneau()


    def _affiche(self):
        with BusyCursor():
            self.dessiner_tableau()

    def naturel(self, m=None):
        u'''Retourne un entier entre 2 et 15.'''
        return str(randint(2, 15))

    def relatif(self, m=None):
        u'''Retourne un entier entre -15 et -2, ou entre 2 et 15.'''
        # signe: 1 ou -1
        signe = 2*randint(0, 1) - 1
        return str(signe*randint(2, 15))

    def decimal(self, m=None):
        u'''Retourne un nombre décimal à deux chiffres.'''
        return self.relatif() + '.' + self.naturel()

    def rationnel(self, m=None):
        u'''Retourne un quotient d'entiers.'''
        while True:
            p = randint(2, 7)
            q = randint(2, 7)
            if p%q:
                break
        signe = 2*randint(0, 1) - 1
        return str(S(signe*p)/S(q))

    def _formater(self, expression):
        expression = expression.replace('**', '^').replace('*', '').replace(' ', '')
        if not re.match('-?(%s)?x?(\(.+\)(\^%s)?)?$' % (NBR, NBR), expression):
            expression = '(' + expression + ')'
        return expression

    def generer_expression(self, expr=None):
        u"""Génère une expression aléatoire en fonction respectant le format
        en cours.

        Si `expr` a une valeur, l'expression reprend la valeur de `expr`.
        """
        # Génération de l'expression:
        x = S('x')
        k = 0
        while True:
            k += 1
            # 10000 essais au maximum
            assert k < 10000
            if expr is None:
                expression = re.sub('n', self.naturel, self.pattern)
                expression = re.sub('z', self.relatif, expression)
                expression = re.sub('d', self.decimal, expression)
                expression = re.sub('q', self.rationnel, expression)
            else:
                expression = expr
            self.raw_expression = expression
            expression = expression.replace('+-', '-').replace('-+', '-')
            if '|' in expression:
                num, den = expression.split('|')
            else:
                num = expression
                den = ''
            num = num.strip()
            den = den.strip()
            if not num or num == '1':
                self.numerateur = []
            else:
                self.numerateur = num.split(',')
            if not den or den == '1':
                self.denominateur = []
            else:
                self.denominateur = den.split(',')
            if not any(gcd(S(P), S(Q)).has(x) for P in self.numerateur for Q in self.denominateur):
                # Il ne faut pas qu'un facteur apparaisse à la fois
                # au numérateur et au dénominateur.
                break

        # Génération de l'expression:
        if len(self.numerateur) > 1:
            num = ''.join(self._formater(facteur) for facteur in self.numerateur)
        else:
            num = self.numerateur[0] if self.numerateur else '1'
        if len(self.denominateur) > 1:
            den = ''.join(self._formater(facteur) for facteur in self.denominateur)
        else:
            den = self.denominateur[0] if self.denominateur else '1'
        if den == '1':
            self.expression = num
        else:
            self.expression = '(%s)/(%s)' %(num, den)

        # On génère tous les dictionnaires utiles à la construction du tableau
        self.expression_latex = convertir_en_latex(self.expression, mode=None)
        num = '*'.join('(' + s + ')' for s in self.numerateur)
        den = '*'.join('(' + s + ')' for s in self.denominateur)
        if den and num:
            self.expression_sympy = S('(%s)/(%s)' %(num, den))
        elif num:
            self.expression_sympy = S('(%s)' %num)
        else:
            assert den
            self.expression_sympy = S('1/(%s)' %den)

        self.facteurs_latex = fact_latex = OrderedDict((expr, convertir_en_latex(expr, mode=None))
                                for expr in self.numerateur + self.denominateur)
        self.facteurs_sympy = fact_sympy = OrderedDict((expr, S(expr)) for expr in fact_latex)
        self.facteurs_sols = OrderedDict((expr, solve(fact_sympy[expr])) for expr in fact_latex)
        self.facteurs_diff = OrderedDict((expr, fact_sympy[expr].diff(S('x'))) for expr in fact_latex)

        # valeurs remarquables de x
        self.sols = []
        for sols in self.facteurs_sols.values():
            self.sols.extend(sols)
        self.sols.sort()
        if param.debug:
            print('(Exercice tableau de signes) Liste des solutions: ' + str(self.sols))



    def dessiner_tableau(self, y=0):
        ##self.fermer_feuille()
        can = self.canvas
        expression_latex = self.expression_latex
        expression_sympy = self.expression_sympy
        facteurs_latex = self.facteurs_latex
        facteurs_sympy = self.facteurs_sympy
        facteurs_sols = self.facteurs_sols
        facteurs_diff = self.facteurs_diff
        sols = self.sols


        # Lorsque l'affichage est actualisé (fenêtre redimensionnée par
        # exemple), le champ n'est pas récréé, mais seulement mis à jour.
        # C'est primordial, car sinon, de nouveaux champs vides se superposent
        # aux anciens, au lieu que les anciens soient déplacés.
        # Si la feuille contient déjà des champs, c'est qu'il s'agit d'une
        # simple mise à jour ; il ne faut donc *pas* recréer les champs.
        creation = self.feuille_actuelle.objets.lister(type=Champ) == []

        # Les pixels et les unités correspondent:
        width, height = can.dimensions
        can.fenetre = 0, width, 0, height

        def dessiner_ligne_h(y, **kw):
            return can.dessiner_ligne((10, width - 30), (height - y, height - y), 'k', **kw)

        def dessiner_champ(x, y, resultat, compteur=[0], **kw):
            va = kw.pop('va', 'center')
            ha = kw.pop('ha', 'center')
            size = kw.pop('size', 16)
            ##Cls = Choix if 'choix' in kw else Champ
            nom = 'champ' + str(compteur[0])
            if creation:
                t = Champ('', x, height - y, couleur='b',
                       alignement_horizontal=ha, alignement_vertical=va,
                       attendu=resultat, taille=size, **kw)
                self.feuille_actuelle.objets[nom] = t
            else:
                t = self.feuille_actuelle.objets[nom]
                t.x = x
                t.y = height - y
                t.style(couleur='b',
                       alignement_horizontal=ha, alignement_vertical=va,
                       attendu=resultat, taille=size, **kw)
            t.evt_valider = self.compter_points
            t.valider = self._valider
            # Compteur qui s'incrémente à chaque nouveau champ.
            # Le but est que chaque champ ait un identifiant unique.
            compteur[0] += 1


        def dessiner_texte(x, y, texte, **kw):
            va = kw.pop('va', 'center')
            size = kw.pop('size', 16)
            return can.dessiner_texte(x, height - y, texte, size=size, va=va, **kw)


        # Paramètre d'espacement (marge entre le bord d'une case et le texte).
        marge = 4

        # --------
        # Consigne
        # --------

        txt = dessiner_texte(10, 10, u"Résoudre sur $\u211D$ l'inéquation suivante :  $"
                                  + expression_latex + u'$.',
                                  va='top', weight='bold', backgroundcolor='#ffffb5')
        box = can.txt_box(txt)
        h = 10 + box.height + 4*marge

        # -------------------------------
        # Équations préalables au tableau
        # -------------------------------

        choix = [u'décroissante', u'croissante']

        # On écrit au dessus du tableau les équations à résoudre :
        for expression, latex in facteurs_latex.items():
            # S'il y a des solutions:
            if facteurs_sols[expression]:
                assert len(facteurs_sols[expression]) == 1, \
                    ("%s a plusieurs solution (non pris en charge pour l'instant)." % expression)
                if facteurs_diff[expression] == 0:
                    continue # Fonction constante

                txt = dessiner_texte(10, h, r'$\bullet\,' + latex +
                                        r'\,=\,0\,\,\Longleftrightarrow\,\,x\,=\,$')
                box = can.txt_box(txt)
                resultat = nice_str(facteurs_sols[expression][0])
                ##print resultat
                dessiner_champ(18 + box.width, h, ha='left', resultat=resultat)
                ##dessiner_texte(220, height - h, u'\u2713', color='g') # 263A  00D8
                ##dessiner_texte(240, height - h, u'\u2639', color='r') #u'\u26A0'
                h += box.height + 2*marge

                if facteurs_sympy[expression].has('x'):
                    if not facteurs_diff[expression].has('x'):
                        # C'est une fonction affine.
                        txt = dessiner_texte(30, h, u'Sur $\u211D$, la fonction affine'
                                               u' $x\\mapsto %s$ est strictement' %latex)
                        box = can.txt_box(txt)
                        sens = (u'décroissante' if facteurs_diff[expression] < 0 else u'croissante')
                        dessiner_champ(35 + box.width, h, ha='left', choix=choix, resultat=sens)
                        h += box.height + 3*marge
                    elif facteurs_sympy[expression].as_base_exp()[1] == 2:
                        # C'est un carré.
                        txt = dessiner_texte(30, h, u'Sur $\u211D$, un carré est toujours')
                        box = can.txt_box(txt)
                        dessiner_champ(35 + box.width, h, ha='left', resultat='positif')
                        h += box.height + 3*marge



        # -------------------------------------------------------------
        # |     x      | -oo        val1     ...     valn          +oo |
        # -------------------------------------------------------------
        # | facteur 1  |                                               |
        # -------------------------------------------------------------
        # |    ...     |                                               |
        # -------------------------------------------------------------
        # | facteur n  |                                               |
        # -------------------------------------------------------------
        # | Expression |                                               |
        # -------------------------------------------------------------


        # -----------------
        # Tableau de signes
        # -----------------

        # Hauteur de chaque ligne horizontale
        hauteurs = []
        # Objets texte de matplotlib
        txts = []
        # Largeur de la première colonne
        largeur_max = 0

        # Hauteur de la 1ère ligne
        tab_hmin = h

        # On n'affiche pas la dernière ligne si il n'y a qu'un seul
        # "facteur". Par ex, si l'expression est juste 2*x+3.
        print_last_line = self.denominateur or len(self.numerateur) > 1

        textes = ['x'] + facteurs_latex.values()
        if print_last_line:
            textes.append(expression_latex)

        # on dessine les lignes verticales et les textes de la 1ère colonne.
        for txt in textes:
            dessiner_ligne_h(h)
            hauteurs.append(h)
            h += marge
            y = height - h
            txt = dessiner_texte(10, y, '$%s$' %txt, ha='center')
            box = can.txt_box(txt)
            hauteur_txt = max(box.height, 25)
            txt.set_y(y - .5*hauteur_txt)
            txts.append(txt)

            h += hauteur_txt + marge
            largeur_max = max(largeur_max, box.width)
        dessiner_ligne_h(h)
        hauteurs.append(h)

        # Hauteur de la 1ère ligne
        tab_hmax = h

        # On centre les textes dans la colonne
        for txt in txts:
            txt.set_x(10 + marge + largeur_max/2)

        def dessiner_ligne_v(x, **kw):
            return can.dessiner_ligne((x, x), (height - tab_hmin, height - tab_hmax) , 'k', **kw)

        # Bords verticaux du tableau
        dessiner_ligne_v(10)
        dessiner_ligne_v(width - 30)
        # Trait vertical à la fin de la première colonne
        col1 = largeur_max + 10 + 2*marge
        dessiner_ligne_v(col1)

        # Position de chaque ligne verticale
        colonnes = [10, col1]

        # On dessine les cases
        n = len(sols)
        largeur_restante = width - 30 - col1 - 2*marge
        largeur_case = largeur_restante/(n + 1)
        choix = [' ', '0', u'\u2551']
        for i, sol in enumerate(sols):
            x = col1 + (i + 1)*largeur_case
            dessiner_ligne_v(x, alpha=.15)
            colonnes.append(x)
            resultat = nice_str(sol)
            dessiner_champ(x, .5*(hauteurs[0] + hauteurs[1]), resultat=resultat)
            resultat_final = '0'
            for k, expr in enumerate(facteurs_sols):
                if sol in facteurs_sols[expr]:
                    resultat = '0'
                    if expr in self.denominateur:
                        resultat_final = u'\u2551'
                else:
                    resultat = ' '
                ##signe = is sol - 1
                h = .5*(hauteurs[k + 1] + hauteurs[k + 2])
                dessiner_champ(x, h, resultat=resultat, choix=choix)
            if print_last_line:
                h = .5*(hauteurs[-1] + hauteurs[-2])
                dessiner_champ(x, h, resultat=resultat_final, choix=choix)

        colonnes.append(width - 30)

        # On place -oo et +oo
        h = .5*(hauteurs[0] + hauteurs[1])
        dessiner_texte(col1 + marge, h, '$-\\infty$')
        dessiner_texte(width - 30 - marge, h, '$+\\infty$', ha='right')

        # Pour remplir le tableau de signes, le plus simple est encore
        # de tester le signe dans chaque case, par exemple avec la valeur
        # centrale de la case.
        xmin = sols[0] - 10   # pour remplacer -oo
        xmax = sols[-1] + 10  # pour remplacer +oo
        valeurs = [xmin] + sols + [xmax]

        var_x = S('x')
        choix = ['-', '+']
        for i, val in enumerate(valeurs[:-1]):
            m = .5*(val + valeurs[i + 1])
            for k, expr in enumerate(facteurs_sols):
                img = facteurs_sympy[expr].subs(var_x, m)
                assert img != 0
                signe = ('+' if img > 0 else '-')
                h = .5*(hauteurs[k + 1] + hauteurs[k + 2])
                x = .5*(colonnes[i + 1] + colonnes[i + 2])
                dessiner_champ(x, h, resultat=signe, choix=choix)

            if print_last_line:
                img = expression_sympy.subs(var_x, m)
                signe = ('+' if img > 0 else '-')
                h = .5*(hauteurs[-2] + hauteurs[-1])
                dessiner_champ(x, h, resultat=signe, choix=choix)

        ##self.feuille_actuelle.interprete.commande_executee()
        ##self.feuille_actuelle.objets._.encadrer('r')

    def autocompleter(self):
        for t in self.feuille_actuelle.objets.lister(type=Champ):
            t.texte = t.style('attendu', color='g')
        self.feuille_actuelle.interprete.commande_executee()

    a = property(autocompleter)

    @staticmethod
    def _valider(reponse, attendu):
        attendu = attendu.replace(' ', '').replace(',', '.')
        reponse = reponse.replace(' ', '').replace(',', '.')
        if attendu == reponse:
            return True
        else:
            try:
                if abs(float(S(attendu) - S(reponse))) < param.tolerance:
                    return True
            except (SympifyError, ValueError):
                pass
            except Exception:
                print_error()
        return False
##
    ##def _valider_txt(reponse, attendu):
        ##return reponse.replace(' ', '').rstrip('.') == attendu.replace(' ', '')

    @staticmethod
    def _calcul_effectue(expr):
        expr = expr.replace(',', '.')
        return bool(re.match("(%s|%s/%s)$" % (NBR_SIGNE, NBR_SIGNE, NBR_SIGNE), expr))

    def compter_points(self, **kw):
        if 'correct' in kw and 'correct_old' in kw and 'champ' in kw:
            champ = kw['champ']
            if kw['correct']:
                if not kw['correct_old']:
                    if not champ.style('choix'):
                        # C'est plus dur s'il n'y a pas de choix proposé.
                        # On accorde une bonification si le résultat est
                        # un minimum simplifié.
                        if self._calcul_effectue(champ.label()):
                            self.score += 3
                        self.score += 1
                    self.score += 1
            else:
                self.score -= 1
                self.erreurs += 1
        if all(obj.correct for obj in self.feuille_actuelle.objets.lister(type=Champ)):
            self.score += 10*(self.niveau + 1)
        self.update_panneau()



# -*- coding: utf-8 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------##
#              WxGeometrie               #
#          fenetre principale            #
##--------------------------------------##
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2013  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys, os, thread, traceback

from PyQt4.QtGui import (QMainWindow, QApplication, QPlainTextEdit, QIcon, QColor, QPalette,
                        QLabel, QWidget, QVBoxLayout, QMessageBox, QTextCursor)
from PyQt4.QtCore import QSize, Qt, pyqtSignal

from ..pylib import uu, print_error, path2, debug, warning
from ..API.console import Console
from ..API.sauvegarde import FichierSession
from ..API.parametres import sauvegarder_module
from .ligne_commande import LigneCommande
from .gestion_session import GestionnaireSession
from .onglets import Onglets
from .. import param
NOMPROG = param.NOMPROG


class PyOnDemandOutputWindow(QPlainTextEdit):
    def __init__(self, title):
        QPlainTextEdit.__init__(self)
        self.setWindowTitle(title)
        self.setReadOnly(True) # assuming QPlainTextEdit
        self.hide()
        self.resize(QSize(450, 300))

    def write(self, s):
        self.show()
        self.moveCursor(QTextCursor.End)
        self.insertPlainText(s.decode(param.encodage)) # again assuming QPlainTextEdit



class FenetrePrincipale(QMainWindow):

    def __init__(self, app, fichier_log=None):
        QMainWindow.__init__(self, None)
        self.setWindowTitle(NOMPROG)
        self.setStyleSheet(".QWidget {background:white}")
#        palette = QPalette()
#        white = QColor(Qt.white)
#        palette.setColor(QPalette.Window, white)
#        palette.setColor(QPalette.AlternateBase, white)
#        self.setPalette(palette)

        self.application = app # pour acceder a l'application en interne
        # Pour que toutes les fenêtres puissent simplement retrouver la
        # fenêtre principale:
        app.fenetre_principale = self


        # À créer avant les onglets
        self.fenetre_sortie = PyOnDemandOutputWindow(title=u"%s - messages."
                                                             % NOMPROG)
        self.fenetre_sortie.setWindowIcon(
                            QIcon(path2(u"%/wxgeometrie/images/icone_log.svg")))
        self.fichier_log = fichier_log

        self.setWindowIcon(QIcon(path2(u"%/wxgeometrie/images/icone.svg")))
        self.barre = self.statusBar()
        self.barre_dte = QLabel(self)
#        self.barre_dte.setText(
        self.barre.addPermanentWidget(self.barre_dte)



        self.message(u"  Bienvenue !", 1)
        self.message(NOMPROG + u" version " + param.version)

        #Ligne de commande de débogage
        self.ligne_commande = LigneCommande(self, 300, action=self.executer_commande,
                        afficher_bouton=False, legende='Ligne de commande :')
        self.ligne_commande.setVisible(param.ligne_commande)

        # Creation des onglets et de leur contenu
        self.onglets = Onglets(self)

        self.mainWidget=QWidget(self) # dummy widget to contain the
                                      # layout manager
        self.setCentralWidget(self.mainWidget)
        self.mainLayout = QVBoxLayout(self.mainWidget)
        self.mainLayout.setSpacing(0)
        self.mainLayout.addWidget(self.ligne_commande, 0)
        self.mainLayout.addWidget(self.onglets)

        self.setMinimumSize(*param.dimensions_fenetre)

        self.console = Console(self)

        self.setAcceptDrops(True)
        self.setFocus()

        # closing == True si l'application est en train d'être fermée
        self.closing = False

        self.gestion = GestionnaireSession(self.onglets)


    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            if any(url.path().endswith('.geo') or url.path().endswith('.geoz')
                    for url in event.mimeData().urls()):
                event.accept()
            else:
                event.ignore()
        else:
            event.ignore()

    def dropEvent(self, event):
        assert event.mimeData().hasUrls()
        for url in event.mimeData().urls():
            if url.path().endswith('.geo') or url.path().endswith('.geoz'):
                self.onglets.ouvrir(url.toLocalFile())

    def afficher_ligne_commande(self, afficher=None):
        u"Afficher ou non la ligne de commande."
        if afficher is not None:
            if isinstance(afficher, bool):
                param.ligne_commande = afficher
            else:
                param.ligne_commande = not param.ligne_commande
            self.ligne_commande.setVisible(param.ligne_commande)
            if param.ligne_commande:
                self.ligne_commande.setFocus()
        return param.ligne_commande


    def mode_debug(self, debug=None):
        u"Passer en mode déboguage."
        if debug is not None:
            if isinstance(debug, bool):
                param.debug = debug
            else:
                param.debug = not param.debug
        if not param.debug:
            self.fenetre_sortie.close()
        return param.debug


    def message(self, texte, lieu=0):
        if lieu == 0:
            self.barre.showMessage(texte)
        else:
            assert lieu == 1
            self.barre_dte.setText(texte)


    def titre(self, texte=None):
        titre = NOMPROG
        if texte:
            titre += ' - ' + uu(texte)
        self.setWindowTitle(titre)


    def executer_commande(self, commande, **kw):
        try:
            self.console.executer(commande)
            self.message(u"Commande interne exécutée.")
            self.ligne_commande.clear()
        except Exception:
            self.message(u"Commande incorrecte.")
            if param.debug:
                raise

    def plein_ecran(self):
        u"Bascule en mode plein écran <-> mode normal."
        self.setWindowState(self.windowState()^Qt.WindowFullScreen)

    def show(self):
        QMainWindow.show(self)
        # Si la liste des onglets n'est pas vide, on indique que l'onglet
        # affiché est devenu onglet actif (on invoque manuellement
        # l'action .evt_changer() qui est associée à chaque changement d'onglet).
        # NOTA: à effectuer **après** QMainWindow.show() !
        # Sinon, par exemple, le champ de saisie de la calculatrice ne peut
        # récupérer le focus à cette occasion.
        if self.onglets.onglet_actuel is not None:
            self.onglets.evt_changer(self.onglets.currentIndex())

    def closeEvent(self, event):
        self.activateWindow()
        self.showNormal()
        self.raise_()
        self.closing = True
        if not param.fermeture_instantanee: # pour des tests rapides
            try:
                if param.confirmer_quitter and not param._restart:
                    panel = self.onglets.onglet_actuel
                    test = hasattr(panel, 'canvas') and hasattr(panel.canvas, 'setUpdatesEnabled')
                    if test:
                        panel.canvas.setUpdatesEnabled(False)
                    reponse = QMessageBox.question(self, u'Quitter %s ?' %NOMPROG,
                                               u'Voulez-vous quitter %s ?' %NOMPROG,
                                               QMessageBox.Yes | QMessageBox.No,
                                               QMessageBox.Yes)
                    if test:
                        panel.canvas.setUpdatesEnabled(True)
                    if reponse != QMessageBox.Yes:
                        self.closing = False
                        event.ignore()
                        return

                self.gestion.fermer()

#                for onglet in self.onglets:
#                    try:
#                        if isinstance(onglet, Panel_API_graphique):
#                            if param.historique_log:
#                                onglet.log.archiver()
#                            onglet.fermer_feuilles()
#                    except:
#                        #print_error()
#                        debug(u"Fermeture incorrecte de l'onglet : ", uu(str(onglet)))
#                        raise

            except Exception:
                try:
                    print_error()
                    QMessageBox.warning(self, u"Erreur lors de la fermeture du programme", traceback.format_exc())
                except UnicodeError:
                    QMessageBox.warning(self, u"Erreur lors de la fermeture du programme", "Impossible d'afficher l'erreur.")
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        if hasattr(self, "fenetre_sortie"):
            self.fenetre_sortie.close()
        print "On ferme !"


    def restart(self):
        param._restart = True
        self.close()
